SystemOrganization addCategory: #'System-Text'!

Object subclass: #Editor
	instanceVariableNames: 'morph selectionShowing'
	classVariableNames: 'CmdKeysInText'
	poolDictionaries: ''
	category: 'System-Text'!

!Editor commentStamp: '<historical>' prior: 0!
New text editors.
TextEditor provides most of the functionality that used to be in TextMorphEditor. This class is no longer a Controller!!
SmalltalkEditor is has Smalltalk code specific features.
SimpleEditor provides basic functionality for single line text editing. It does not handle fonts and styles, aligning and Smalltalk utilities. It handles one single line.
CellStyleEditor allows entering alphabetic characters using only number keys, like most cell phones do.!

!Editor class methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 12:40'!
cmdKeysInText
	^ CmdKeysInText ifNil: [CmdKeysInText := true]! !

!Editor class methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 12:40'!
cmdKeysInText: aBoolean
	CmdKeysInText := aBoolean! !

!Editor class methodsFor: 'private accessing' stamp: 'FernandoOlivero 3/16/2010 00:25'!
specialShiftCmdKeys

"Private - return array of key codes that represent single keys acting
as if shift-command were also being pressed"

^#(
	1	"home"
	3	"enter"
	4	"end"
	8	"backspace"
	11	"page up"
	12	"page down"
	27	"escape"
	28	"left arrow"
	29	"right arrow"
	30	"up arrow"
	31	"down arrow"
	127	"delete"
	)! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:37'!
backWord: aKeyboardEvent
	"If the selection is not a caret, delete it and leave it in the backspace buffer.
	 Else, delete the word before the caret."

	| startIndex |
	self hasCaret ifTrue: [ "a caret, delete at least one character"
		startIndex := 1 max: self markIndex - 1.
		[startIndex > 1 and:
			[(self string at: startIndex - 1) tokenish]]
				whileTrue: [
					startIndex := startIndex - 1]]
	ifFalse: [ "a non-caret, just delete it"
		startIndex := self markIndex].
	self backTo: startIndex.
	^false! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 13:01'!
backspace: aKeyboardEvent
	"Backspace over the last character."

	| startIndex |
	aKeyboardEvent shiftPressed ifTrue: [^ self backWord: aKeyboardEvent keyCharacter].
	startIndex := self markIndex + (self hasCaret ifTrue: [0] ifFalse: [1]).
	startIndex := 1 max: startIndex - 1.
	self backTo: startIndex.
	^false! !

!Editor methodsFor: 'menu messages' stamp: 'jmv 11/5/2008 10:16'!
clipboardText

	^ Clipboard clipboardText! !

!Editor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 23:42'!
clipboardTextPut: text

	^ Clipboard clipboardText: text! !

!Editor methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 12:43'!
cmdKeysInText
	^ self class cmdKeysInText! !

!Editor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:37'!
copySelection: aKeyboardEvent
	"Copy the current text selection."

	self copySelection.
	^true! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:37'!
crlf: aKeyboardEvent
	"Append a line feed character to the stream of characters."

	self addString: String crlf.
	^false! !

!Editor methodsFor: 'accessing' stamp: 'jmv 9/1/2009 18:09'!
currentAttributes
	"Redefined by subclasses that handle TextAttributes"
	^nil! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 13:17'!
cursorDown: aKeyboardEvent 

	"Private - Move cursor from position in current line to same position in
	next line. If next line too short, put at end. If shift key down,
	select."

	self closeTypeIn.
	self 
		moveCursor:[:position | self
				sameColumn: position
				newLine:[:line | line + 1]
				forward: true]
		forward: true
		specialBlock:[:dummy | dummy]
		event: aKeyboardEvent.
	^true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 13:17'!
cursorLeft: aKeyboardEvent 
	"Private - Move cursor left one character if nothing selected, otherwise 
	move cursor to beginning of selection. If the shift key is down, start 
	selecting or extending current selection. Don't allow cursor past 
	beginning of text"

	self closeTypeIn.
	self
		moveCursor:[:position | position - 1 max: 1]
		forward: false
		specialBlock:[:position | self previousWord: position]
		event: aKeyboardEvent.
	^ true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 13:17'!
cursorPageDown: aKeyboardEvent 

	self closeTypeIn.
	self 
		moveCursor: [:position |
			self
				sameColumn: position
				newLine: [:lineNo | lineNo + self pageHeight]
				forward: true]
		forward: true
		specialBlock:[:dummy | dummy]
		event: aKeyboardEvent.
	^true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 13:17'!
cursorPageUp: aKeyboardEvent 

	self closeTypeIn.
	self 
		moveCursor: [:position |
			self
				sameColumn: position
				newLine: [:lineNo | lineNo - self pageHeight]
				forward: false]
		forward: false
		specialBlock:[:dummy | dummy]
		event: aKeyboardEvent.
	^true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 13:17'!
cursorRight: aKeyboardEvent 
	"Private - Move cursor right one character if nothing selected, 
	otherwise move cursor to end of selection. If the shift key is down, 
	start selecting characters or extending already selected characters. 
	Don't allow cursor past end of text"

	self closeTypeIn.
	self
		moveCursor: [:position | position + 1]
		forward: true
		specialBlock:[:position | self nextWord: position]
		event: aKeyboardEvent.
	^ true! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:37'!
cursorTopHome: aKeyboardEvent
	"Put cursor at beginning of text -- invoked from cmd-H shortcut, useful for keyboards that have no home key."

	self selectAt: 1.
	^ true! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 13:17'!
cursorUp: aKeyboardEvent 
"Private - Move cursor from position in current line to same position in
prior line. If prior line too short, put at end"

	self closeTypeIn.
	self
		moveCursor: [:position | self
				sameColumn: position
				newLine:[:line | line - 1]
				forward: false]
		forward: false
		specialBlock:[:dummy | dummy]
		event: aKeyboardEvent.
	^true! !

!Editor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:38'!
cut: aKeyboardEvent 
	"Cut out the current text selection."

	self cut.
	^true! !

!Editor methodsFor: 'current selection' stamp: 'jmv 11/4/2008 13:20'!
deselect
	"If the text selection is visible on the screen, reverse its highlight."

	selectionShowing ifTrue: [self reverseSelection]! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'FernandoOlivero 3/16/2010 01:01'!
escape 
	self morph escapePressed .
	! !

!Editor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 23:43'!
hasSelection
	^self hasCaret not! !

!Editor methodsFor: 'testing' stamp: 'FernandoOlivero 6/15/2010 13:08'!
isSimpleEditor 
	^ false ! !

!Editor methodsFor: 'testing' stamp: 'FernandoOlivero 6/15/2010 13:07'!
isSmalltalkEditor
	^ false ! !

!Editor methodsFor: 'testing' stamp: 'FernandoOlivero 6/15/2010 13:07'!
isTextEditor
	^ false ! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:38'!
lf: aKeyboardEvent 
	"Append a line feed character to the stream of characters."

	self addString: Character lf asString.
	^false! !

!Editor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 23:42'!
lineSelectAndEmptyCheck: returnBlock
	"If the current selection is an insertion point, expand it to be the entire current line; if after that's done the selection is still empty, then evaluate the returnBlock, which will typically consist of '[^ self]' in the caller -- check senders of this method to understand this."

	self selectLine.  "if current selection is an insertion point, then first select the entire line in which occurs before proceeding"
	self hasSelection ifFalse: [morph flash.  ^ returnBlock value]! !

!Editor methodsFor: 'private' stamp: 'jmv 11/5/2008 10:04'!
lines
	"Compute lines based on logical line breaks, not optical (which may change due to line wrapping of the editor).
	Subclasses using kinds of Paragraphs can instead use the service provided by it.
	"
	| lines string index lineIndex stringSize |
	string := self string.
	"Empty strings have no lines at all. Think of something."
	string isEmpty ifTrue:[^{#(1 0 0)}].
	stringSize := string size.
	lines := OrderedCollection new: (string size // 15).
	index := 0.
	lineIndex := 0.
	string linesDo:[:line |
		lines addLast: (Array
			with: (index := index + 1)
			with: (lineIndex := lineIndex + 1)
			with: (index := index + line size min: stringSize))].
	"Special workaround for last line empty."
	string last == Character cr
	"lines last last < stringSize" ifTrue:[lines addLast:{stringSize +1. lineIndex+1. stringSize}].
	^lines! !

!Editor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 13:06'!
morph
	^ morph! !

!Editor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 13:06'!
morph: aMorph
	"Install a link back to the morph being edited (esp for text links)"
	morph := aMorph ! !

!Editor methodsFor: 'private' stamp: 'jmv 9/2/2009 13:18'!
moveCursor: directionBlock forward: forward specialBlock: specialBlock event: aKeyboardEvent
	"Private - Move cursor.
	directionBlock is a one argument Block that computes the new Position from a given one.
	specialBlock is a one argumentBlock that computes the new position from a given one under the alternate semantics.
	Note that directionBlock always is evaluated first."
	| shift indices newPosition |
	shift := aKeyboardEvent shiftPressed.
	indices := self setIndices: shift forward: forward.
	newPosition := directionBlock value: (indices at: #moving).
	(aKeyboardEvent commandKeyPressed or: [ aKeyboardEvent controlKeyPressed ])
		ifTrue: [newPosition := specialBlock value: newPosition].
	shift
		ifTrue: [self selectMark: (indices at: #fixed) point: newPosition - 1]
		ifFalse: [self selectAt: newPosition]! !

!Editor methodsFor: 'private' stamp: 'jmv 11/4/2008 23:39'!
nextWord: position
	| string index |
	string := self string.
	index := position.
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric]]
		whileTrue: [index := index + 1].
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric not]]
		whileTrue: [index := index + 1].
	^ index! !

!Editor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:38'!
noop: aKeyboardEvent 
	"Unimplemented keyboard command; just ignore it."

	^ true! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'FernandoOlivero 5/17/2010 13:54'!
normalCharacter: aKeyboardEvent 
	"A nonspecial character is to be added to the stream of characters."

	self addString: aKeyboardEvent keyCharacter asString.
	^false! !

!Editor methodsFor: 'menu messages' stamp: 'jmv 11/5/2008 10:12'!
paste
	"Paste the text from the shared buffer over the current selection and 
	redisplay if necessary.  Undoer & Redoer: undoAndReselect."

	self replace: self selectionInterval with: self clipboardText and:
		[self selectAt: self pointIndex]! !

!Editor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:38'!
paste: aKeyboardEvent 
	"Replace the current text selection by the text in the shared buffer."

	self closeTypeIn.
	self paste.
	^true! !

!Editor methodsFor: 'private' stamp: 'jmv 11/4/2008 23:02'!
previousWord: position
	| string index |
	string := self string.
	index := position.
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric not]]
		whileTrue: [index := index - 1].
	[(index between: 1 and: string size) and: [(string at: index) isAlphaNumeric]]
		whileTrue: [index := index - 1].
	^ index + 1! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 11/4/2008 23:48'!
selectAll

	self selectFrom: 1 to: self string size! !

!Editor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:38'!
selectAll: aKeyboardEvent 
	"select everything, invoked by cmd-a.  1/17/96 sw"

	self closeTypeIn.
	self selectFrom: 1 to: self string size.
	^ true! !

!Editor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:10'!
selectAt: characterIndex 
	"Deselect, then place the caret before the character at characterIndex.
	 Be sure it is in view."

	self selectFrom: characterIndex to: characterIndex - 1! !

!Editor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:49'!
selectInterval: anInterval
	"Deselect, then select the specified characters inclusive.
	 Be sure the selection is in view."

	self selectFrom: anInterval first to: anInterval last! !

!Editor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:13'!
selectInvisiblyFrom: start to: stop
	"Select the designated characters, inclusive.  Make no visual changes."

	self markIndex: start; pointIndex: stop + 1! !

!Editor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:14'!
selectInvisiblyMark: mark point: point
	"Select the designated characters, inclusive.  Make no visual changes."

	self markIndex: mark; pointIndex: point + 1! !

!Editor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:13'!
selectMark: mark point: point
	"Deselect, then select the specified characters inclusive.
	 Be sure the selection is in view."

	(mark =  self markIndex and: [point + 1 = self pointIndex]) ifFalse:
		[self deselect.
		self selectInvisiblyMark: mark point: point]! !

!Editor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:36'!
selectWord
	"Select delimited text or word--the result of double-clicking."

	| openDelimiter closeDelimiter direction match level leftDelimiters rightDelimiters
	string here hereChar start stop |
	string := self string.
	here := self pointIndex.
	(here between: 2 and: string size)
		ifFalse: ["if at beginning or end, select entire string"
			^self selectFrom: 1 to: string size].
	leftDelimiters := '([{<''"
'.
	rightDelimiters := ')]}>''"
'.
	openDelimiter := string at: here - 1.
	match := leftDelimiters indexOf: openDelimiter.
	match > 0
		ifTrue: 
			["delimiter is on left -- match to the right"
			start := here.
			direction := 1.
			here := here - 1.
			closeDelimiter := rightDelimiters at: match]
		ifFalse: 
			[openDelimiter := string at: here.
			match := rightDelimiters indexOf: openDelimiter.
			match > 0
				ifTrue: 
					["delimiter is on right -- match to the left"
					stop := here - 1.
					direction := -1.
					closeDelimiter := leftDelimiters at: match]
				ifFalse: ["no delimiters -- select a token"
					direction := -1]].
	level := 1.
	[level > 0 and: [direction > 0
			ifTrue: [here < string size]
			ifFalse: [here > 1]]]
		whileTrue: 
			[hereChar := string at: (here := here + direction).
			match = 0
				ifTrue: ["token scan goes left, then right"
					hereChar tokenish
						ifTrue: [here = 1
								ifTrue: 
									[start := 1.
									"go right if hit string start"
									direction := 1]]
						ifFalse: [direction < 0
								ifTrue: 
									[start := here + 1.
									"go right if hit non-token"
									direction := 1]
								ifFalse: [level := 0]]]
				ifFalse: ["bracket match just counts nesting level"
					hereChar = closeDelimiter
						ifTrue: [level := level - 1"leaving nest"]
						ifFalse: [hereChar = openDelimiter 
									ifTrue: [level := level + 1"entering deeper nest"]]]].

	level > 0 ifTrue: ["in case ran off string end"	here := here + direction].
	direction > 0
		ifTrue: [self selectFrom: start to: here - 1]
		ifFalse: [self selectFrom: here + 1 to: stop]! !

!Editor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 12:38'!
selectWord: aKeyboardEvent

	self closeTypeIn.
	self selectWord.
	^ true! !

!Editor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:24'!
selectionInterval
	"Answer the interval that is currently selected."

	^self startIndex to: self stopIndex - 1 ! !

!Editor methodsFor: 'private' stamp: 'AlainPlantec 4/15/2010 11:41'!
setIndices: shiftPressed forward: forward
	"Little helper method that sets the moving and fixed indices according to some flags."
	| indices |
	indices := Dictionary new.
	self flag: 'to be reviewed'.
	(shiftPressed and: [false "Preferences selectionsMayShrink"])
		ifTrue: [
			indices at: #moving put: self pointIndex.
			indices at: #fixed put: self markIndex
		] ifFalse: [
			forward
				ifTrue:[
					indices at: #moving put: self stopIndex.
					indices at: #fixed put: self startIndex.
				] ifFalse: [
					indices at: #moving put: self startIndex.
					indices at: #fixed put: self stopIndex.
				]
		].
	^indices! !

!Editor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 14:02'!
unselect
	self markIndex: self pointIndex! !

!Editor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 14:33'!
userHasEdited
	"Note that my text is free of user edits."

	morph hasUnacceptedEdits: true! !

Editor subclass: #SimpleEditor
	instanceVariableNames: 'string pointIndex markIndex beginTypeInIndex'
	classVariableNames: 'CmdActions ShiftCmdActions'
	poolDictionaries: ''
	category: 'System-Text'!

!SimpleEditor commentStamp: '<historical>' prior: 0!
See comment in Editor!

!SimpleEditor class methodsFor: 'as yet unclassified' stamp: 'FernandoOlivero 3/16/2010 00:26'!
initialize 
	"Initialize the keyboard shortcut maps and the shared buffers
	for copying text across views and managing again and undo." 
 
	"
	SimpleEditor initialize
	"

	self initializeCmdKeyShortcuts.
	self initializeShiftCmdKeyShortcuts! !

!SimpleEditor class methodsFor: 'as yet unclassified' stamp: 'FernandoOlivero 3/16/2010 00:30'!
initializeCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table."

	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"

	"SimpleTextEditor initialize"

	| cmdMap cmds |
	cmdMap := Array new: 256 withAll: #noop:.	"use temp in case of a crash"
	cmdMap at: 1 + 1 put: #cursorHome:.	"home key"
	cmdMap at: 4 + 1 put: #cursorEnd:.	"end key"
	cmdMap at: 8 + 1 put: #backspace:.	"ctrl-H or delete key"
	cmdMap at: 28 + 1 put: #cursorLeft:.	"left arrow key"
	cmdMap at: 29 + 1 put: #cursorRight:.	"right arrow key"
	cmdMap at: 32 + 1 put: #selectWord:.	"space bar key"
	cmdMap at: 127 + 1 put: #forwardDelete:.	"del key"
	
	cmds := #($a #selectAll: $c #copySelection: $v #paste: $w #backWord: $x #cut: ).
	1 to: cmds size
		by: 2
		do: [:i | cmdMap at: (cmds at: i) asciiValue + 1 put: (cmds at: i + 1)].
	CmdActions := cmdMap! !

!SimpleEditor class methodsFor: 'as yet unclassified' stamp: 'FernandoOlivero 3/29/2010 17:02'!
initializeShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."
	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"
	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	| cmdMap cmds |

	"shift-command and control shortcuts"
	cmdMap := Array new: 256 withAll: #noop:.  "use temp in case of a crash"
	cmdMap at: ( 1 + 1) put: #cursorHome:.			"home key"
	cmdMap at: ( 4 + 1) put: #cursorEnd:.			"end key"
	cmdMap at: ( 8 + 1) put: #forwardDelete:.		"ctrl-H or delete key"
	cmdMap at: (28 + 1) put: #cursorLeft:.			"left arrow key"
	cmdMap at: (29 + 1) put: #cursorRight:.			"right arrow key"
	cmdMap at: (27 + 1) put: #escape.					"escape"				
	cmdMap at: (32 + 1) put: #selectWord:.			"space bar key"
	cmdMap at: (127 + 1) put: #forwardDelete:.		"del key"

	cmds := #(
		$x	makeLowercase:
		$y	makeUppercase:
		$z	makeCapitalized:
	).
	1 to: cmds size by: 2 do: [ :i |
		cmdMap at: ((cmds at: i) asciiValue + 1) put: (cmds at: i + 1).		"plain keys"
		cmdMap at: ((cmds at: i) asciiValue - 32 + 1) put: (cmds at: i + 1).		"shifted keys"
		cmdMap at: ((cmds at: i) asciiValue - 96 + 1) put: (cmds at: i + 1).		"ctrl keys"
	].
	ShiftCmdActions := cmdMap! !

!SimpleEditor methodsFor: 'typing support' stamp: 'jmv 9/2/2009 07:23'!
addString: aString
	"Think of a better name"

	self zapSelectionWith: aString! !

!SimpleEditor methodsFor: 'typing support' stamp: 'jmv 9/2/2009 11:47'!
backTo: startIndex
	"During typing, backspace to startIndex.  Deleted characters fall into three
	 clusters, from left to right in the text: (1) preexisting characters that were
	 backed over; (2) newly typed characters that were backed over; 
	(3) preexisting characters that
	 were highlighted before typing began.  If typing has not yet been opened,
	 open it and watch for the first and third cluster.  If typing has been opened,
	 watch for the first and second cluster.  
	 Delete all the clusters.  The code is shorter than the comment."

	| saveLimit |
	saveLimit := beginTypeInIndex
		ifNil: [self openTypeIn. self stopIndex]
		ifNotNil: [self startOfTyping].
	self markIndex: startIndex.
	startIndex < saveLimit ifTrue: [
		beginTypeInIndex := self startIndex.].
	self zapSelectionWith: ''.
	self unselect! !

!SimpleEditor methodsFor: 'initialize-release' stamp: 'jmv 11/4/2008 13:08'!
changeString: aString
	"Install aString as the one to be edited by the receiver."

	string := aString.
	self resetState! !

!SimpleEditor methodsFor: 'typing support' stamp: 'jmv 9/2/2009 11:47'!
closeTypeIn
	"See comment in openTypeIn.  It is important to call closeTypeIn before executing
	 any non-typing key, making a new selection, etc.  It is called automatically for
	 menu commands."

	beginTypeInIndex := nil! !

!SimpleEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 23:42'!
copySelection
	"Copy the current selection and store it in the paste buffer, unless a caret.  Undoer & Redoer: undoCutCopy"

	self lineSelectAndEmptyCheck: [^ self].
	self clipboardTextPut: self selection! !

!SimpleEditor methodsFor: 'nonediting/nontyping keys' stamp: 'StephaneDucasse 3/31/2010 22:06'!
cursorEnd: aKeyboardEvent

	"Private - Move cursor end of current line."
	self closeTypeIn.
	self
		moveCursor:
			[ :position | 
				string
					indexOf: Character cr
					startingAt: position
					ifAbsent:[string size + 1]]
		forward: true
		specialBlock: [:dummy | string size + 1]
		event: aKeyboardEvent.
	^true! !

!SimpleEditor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 13:18'!
cursorHome: aKeyboardEvent 

	"Private - Move cursor from position in current line to beginning of
	current line. If control key is pressed put cursor at beginning of text"

	self
		moveCursor: [ :position |
				(string
					lastIndexOf: Character cr
					startingAt: position - 1
					ifAbsent:[0]) + 1]
		forward: false
		specialBlock: [:dummy | 1]
		event: aKeyboardEvent.
	^true! !

!SimpleEditor methodsFor: 'menu messages' stamp: 'jmv 11/5/2008 10:21'!
cut
	"Cut out the current selection and redisplay the paragraph if necessary.  Undoer & Redoer: undoCutCopy:"

	self lineSelectAndEmptyCheck: [^ self].
	self replaceSelectionWith: ''. 
	self clipboardTextPut: self selection! !

!SimpleEditor methodsFor: 'typing support' stamp: 'FernandoOlivero 5/17/2010 15:14'!
dispatchOn: aKeyboardEvent
	"Carry out the action associated with this character, if any.
	Type-ahead is passed so some routines can flush or use it."

	| asciiValue honorCommandKeys |

	"We don't support multiple lines. Therefore, we don't process cr as a #normalCharacter:"
	asciiValue := aKeyboardEvent keyValue.
	asciiValue = Character cr asciiValue 
		ifTrue: [
			morph acceptOnCR
				ifTrue:[
					morph acceptContents .
					self closeTypeIn.
					^ true].
			aKeyboardEvent controlKeyPressed ifTrue: [ ^ self normalCharacter: aKeyboardEvent ].
			aKeyboardEvent shiftPressed ifTrue: [ ^ self lf: aKeyboardEvent ].
			aKeyboardEvent commandKeyPressed ifTrue: [ ^ self crlf: aKeyboardEvent ].
			^ self crWithIndent: aKeyboardEvent ].

	honorCommandKeys := self cmdKeysInText.

	"Special keys overwrite crtl+key combinations - at least on Windows. To resolve this
	conflict, assume that keys other than cursor keys aren't used together with Crtl." 
	((self class specialShiftCmdKeys includes: asciiValue) and: [ asciiValue < 27])
		ifTrue: [^ aKeyboardEvent controlKeyPressed
			ifTrue: [self perform: (ShiftCmdActions at: asciiValue + 1) with: aKeyboardEvent]
			ifFalse: [self perform: (CmdActions at: asciiValue + 1) with: aKeyboardEvent]].

	"backspace, and escape keys (ascii 8 and 27) are command keys"
	((honorCommandKeys and: [aKeyboardEvent commandKeyPressed]) or: [self class specialShiftCmdKeys includes: asciiValue]) ifTrue: [
		^ aKeyboardEvent shiftPressed
			ifTrue: [
				self perform: (ShiftCmdActions at: asciiValue + 1) with: aKeyboardEvent]
			ifFalse: [
				self perform: (CmdActions at: asciiValue + 1) with: aKeyboardEvent]].

	"the control key can be used to invoke shift-cmd shortcuts"
	(honorCommandKeys and: [aKeyboardEvent controlKeyPressed])
		ifTrue: [
			^ self perform: (ShiftCmdActions at: asciiValue + 1) with: aKeyboardEvent].

	^ self normalCharacter: aKeyboardEvent! !

!SimpleEditor methodsFor: 'typing support' stamp: 'jmv 11/4/2008 14:04'!
doneTyping
	beginTypeInIndex := nil! !

!SimpleEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 13:02'!
forwardDelete: aKeyboardEvent
	"Delete forward over the next character.
	  Make Undo work on the whole type-in, not just the one char.
	wod 11/3/1998: If there was a selection use #zapSelectionWith: rather than #backspace: which was 'one off' in deleting the selection. Handling of things like undo or typeIn area were not fully considered."

	| startIndex stopIndex |
	startIndex := self markIndex.
	startIndex > string size ifTrue: [
		^ false].
	self hasSelection ifTrue: [
		"there was a selection"
		self zapSelectionWith: ''.
		^ false].
	"Null selection - do the delete forward"
	beginTypeInIndex isNil	"no previous typing.  openTypeIn"
		ifTrue: [self openTypeIn].
	stopIndex := startIndex.
	(aKeyboardEvent keyValue = 127 and: [ aKeyboardEvent shiftPressed ])
		ifTrue: [ stopIndex := (self nextWord: stopIndex) - 1 ].
	self selectFrom: startIndex to: stopIndex.
	self replaceSelectionWith: ''.
	self selectFrom: startIndex to: startIndex-1.
	^false! !

!SimpleEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 23:24'!
hasCaret
	^markIndex = pointIndex! !

!SimpleEditor methodsFor: 'initialize-release' stamp: 'jmv 11/4/2008 13:42'!
initialize
	"Initialize the state of the receiver. Subclasses should include 'super 
	initialize' when redefining this message to insure proper initialization."

	super initialize.
	self changeString: ''! !

!SimpleEditor methodsFor: 'testing' stamp: 'jmv 11/17/2008 17:01'!
isAtEnd
	^pointIndex = (string size + 1)! !

!SimpleEditor methodsFor: 'testing' stamp: 'jmv 11/17/2008 17:01'!
isAtStart
	^pointIndex = 1! !

!SimpleEditor methodsFor: 'testing' stamp: 'FernandoOlivero 6/15/2010 13:07'!
isSimpleEditor 
	^ true ! !

!SimpleEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:40'!
makeCapitalized: aKeyboardEvent
	"Force the current selection to uppercase.  Triggered by Cmd-X."
	| prev |
	prev := $-.  "not a letter"
	self replaceSelectionWith:
		(self selection collect: [:c |
			prev := prev isLetter ifTrue: [c asLowercase] ifFalse: [c asUppercase]]).
	^ true! !

!SimpleEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:40'!
makeLowercase: aKeyboardEvent
	"Force the current selection to lowercase.  Triggered by Cmd-X."

	self replaceSelectionWith: (self selection asLowercase).
	^ true! !

!SimpleEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:40'!
makeUppercase: aKeyboardEvent
	"Force the current selection to uppercase.  Triggered by Cmd-Y."

	self replaceSelectionWith: (self selection asUppercase).
	^ true! !

!SimpleEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:57'!
markIndex
	^markIndex! !

!SimpleEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/5/2008 11:41'!
markIndex: anIndex
	markIndex := anIndex min: string size + 1 max: 1! !

!SimpleEditor methodsFor: 'events' stamp: 'jmv 11/5/2008 13:54'!
mouseDown: evt index: index
	
	self closeTypeIn.
	evt shiftPressed
		ifFalse: [
			self pointIndex: index.
			self markIndex: index]
		ifTrue: [
			self mouseMove: evt ]! !

!SimpleEditor methodsFor: 'events' stamp: 'jmv 11/5/2008 14:10'!
mouseMove: evt index: index
	"Change the selection in response to moue-down drag"

	self pointIndex: index! !

!SimpleEditor methodsFor: 'typing support' stamp: 'jmv 7/30/2009 00:45'!
openTypeIn
	"Set up beginTypeInIndex to keep track of the leftmost backspace."

	beginTypeInIndex ifNil: [
		beginTypeInIndex := self startIndex ]! !

!SimpleEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:57'!
pointIndex
	^ pointIndex! !

!SimpleEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/5/2008 11:41'!
pointIndex: anIndex
	pointIndex := anIndex min: string size + 1 max: 1! !

!SimpleEditor methodsFor: 'typing support' stamp: 'FernandoOlivero 1/21/2010 14:47'!
processKeyStroke: aKeyboardEvent

	"Key struck on the keyboard. Find out which one and, if special, carry 
	out the associated special action. Otherwise, add the character to the 
	stream of characters.  Undoer & Redoer: see closeTypeIn."

		self deselect.
			
			(self dispatchOn: aKeyboardEvent) ifTrue: [
				self doneTyping.
				^self].
			self openTypeIn.

		self unselect! !

!SimpleEditor methodsFor: 'accessing' stamp: 'jmv 11/5/2008 12:22'!
replace: oldInterval with: newString and: selectingBlock 
	"Replace the text in oldInterval with newString and execute selectingBlock to establish the new selection."

	self selectionInterval = oldInterval ifFalse: [self selectInterval: oldInterval].
	self zapSelectionWith: newString.
	selectingBlock value! !

!SimpleEditor methodsFor: 'accessing' stamp: 'jmv 11/5/2008 11:17'!
replaceSelectionWith: aString
	"Deselect, and replace the selection text by aText."

	self zapSelectionWith: aString! !

!SimpleEditor methodsFor: 'initialize-release' stamp: 'jmv 11/4/2008 14:12'!
resetState 
	"Establish the initial conditions for editing the paragraph: place caret 
	before first character, set the emphasis to that of the first character,
	and save the paragraph for purposes of canceling."

	pointIndex := 1.	"Like pointBlock in EnhancedTextEditor"
	markIndex := 1.	"Like markIndex in EnhancedTextEditor"
	beginTypeInIndex := nil.	"Like beginTypeInBlock in EnhancedTextEditor"
	selectionShowing := false! !

!SimpleEditor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:12'!
selectFrom: start to: stop
	"Select the specified characters inclusive."
	self selectInvisiblyFrom: start to: stop.
	self closeTypeIn! !

!SimpleEditor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 23:49'!
selectLine
	"Make the receiver's selection, if it currently consists of an insertion point only, encompass the current line."
	self hasSelection ifTrue:[^self].
	self selectAll! !

!SimpleEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/5/2008 10:24'!
selection

	^string copyFrom: self startIndex to: self stopIndex - 1 ! !

!SimpleEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:55'!
startIndex
	^ pointIndex min: markIndex! !

!SimpleEditor methodsFor: 'typing support' stamp: 'jmv 11/4/2008 13:28'!
startOfTyping
	^ beginTypeInIndex! !

!SimpleEditor methodsFor: 'initialize-release' stamp: 'jmv 11/4/2008 13:23'!
stateArray
	^ {	self selectionInterval.
		self startOfTyping}! !

!SimpleEditor methodsFor: 'initialize-release' stamp: 'jmv 11/4/2008 13:23'!
stateArrayPut: stateArray
	| sel |
	sel := stateArray at: 1.
	self selectFrom: sel first to: sel last.
	beginTypeInIndex := stateArray at: 2! !

!SimpleEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:55'!
stopIndex
	^ pointIndex max: markIndex! !

!SimpleEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 16:09'!
string
	^string! !

!SimpleEditor methodsFor: 'mvc compatibility' stamp: 'jmv 8/19/2009 17:39'!
zapSelectionWith: aString

	| start stop |
	self deselect.
	start := self startIndex.
	stop := self stopIndex.
	(start = stop and: [aString size = 0]) ifFalse: [
		string := string copyReplaceFrom: start to: stop-1 with: aString.
		self
			markIndex: start; 
			pointIndex: start + aString size.

		self userHasEdited  " -- note text now dirty" ]! !

Editor subclass: #TextEditor
	instanceVariableNames: 'model paragraph pointBlock markBlock beginTypeInBlock emphasisHere otherInterval lastParenLocation oldInterval'
	classVariableNames: 'CaseSensitiveFinds ChangeText FindText UndoInterval UndoMessage UndoParagraph UndoSelection Undone WalkAlongDisplayedLine'
	poolDictionaries: ''
	category: 'System-Text'!

TextEditor class
	instanceVariableNames: 'cmdActions shiftCmdActions yellowButtonMenu shiftedYellowButtonMenu'!

!TextEditor commentStamp: '<historical>' prior: 0!
See comment in Editor.

My instances edit Text, this is, they support multiple lines and TextAttributes.
They have no specific facilities for editing Smalltalk code. Those are found in SmalltalkEditor.!

TextEditor class
	instanceVariableNames: 'cmdActions shiftCmdActions yellowButtonMenu shiftedYellowButtonMenu'!

TextEditor subclass: #SmalltalkEditor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Text'!

!SmalltalkEditor commentStamp: '<historical>' prior: 0!
The editor built specifically for Smalltalk code!

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'jmv 8/18/2009 18:21'!
initializeCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table."

	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"

	"SmalltalkEditor initialize"

	| cmds |
	self initializeBasicCmdKeyShortcuts.
	
	cmds := #($b #browseIt: $d #doIt: $i #inspectIt: $j #doAgainOnce: $l #cancel: $m #implementorsOfIt: $n #sendersOfIt: $o #spawnIt: $p #printIt: $q #querySymbol: $s #save: ).
	1 to: cmds size
		by: 2
		do: [:i | cmdActions at: (cmds at: i) asciiValue + 1 put: (cmds at: i + 1)]! !

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'jmv 8/18/2009 18:25'!
initializeShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."
	"NOTE: if you don't know what your keyboard generates, use Sensor kbdTest"
	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	"SmalltalkEditor initialize"

	| cmds |
 	self initializeBasicShiftCmdKeyShortcuts.
	
	cmds := #(
		$a	argAdvance:
		$b	browseItHere:
		$e	methodStringsContainingIt:
		$f	displayIfFalse:
		$g	fileItIn:
		$i	exploreIt:
		$n	referencesToIt:
		$t	displayIfTrue:
		$v	pasteInitials:
		$w	methodNamesContainingIt:
	).
	1 to: cmds size by: 2 do: [ :i |
		shiftCmdActions at: ((cmds at: i) asciiValue + 1) put: (cmds at: i + 1).			"plain keys"
		shiftCmdActions at: ((cmds at: i) asciiValue - 32 + 1) put: (cmds at: i + 1).		"shifted keys"
		shiftCmdActions at: ((cmds at: i) asciiValue - 96 + 1) put: (cmds at: i + 1).		"ctrl keys"
	].! !

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'StephaneDucasse 7/2/2010 18:24'!
initializeShiftedYellowButtonMenu
	"Initialize the yellow button pop-up menu and corresponding messages."

	"SmalltalkEditor initialize"

	shiftedYellowButtonMenu := MenuMorph new  defaultTarget: self.
	shiftedYellowButtonMenu
		addList:
	{
		{'explain' translated.						#explain}.
		{'pretty print' translated.					#prettyPrint}.
		{'file it in (G)' translated.					#fileItIn}.
		{'spawn (o)' translated.						#spawn}.
		#-.
		{'browse it (b)' translated.					#browseIt}.
		{'senders of it (n)' translated.				#sendersOfIt}.
		{'implementors of it (m)' translated.		#implementorsOfIt}.
		{'references to it (N)' translated.			#referencesToIt}.
		#-.
		{'selectors containing it (W)' translated.	#methodNamesContainingIt}.
		{'method strings with it (E)' translated.		#methodStringsContainingit}.
		{'method source with it' translated.		#methodSourceContainingIt}.
		{'class names containing it' translated.		#classNamesContainingIt}.
		{'class comments with it' translated.		#classCommentsContainingIt}.
		{'change sets with it' translated.			#browseChangeSetsWithSelector}.
		#-.
		{'save contents to file...' translated.		#saveContentsInFile}.
		#-.
		{'more...' translated.						#yellowButtonActivity}.
	} ! !

!SmalltalkEditor class methodsFor: 'keyboard shortcut tables' stamp: 'StephaneDucasse 7/2/2010 18:48'!
initializeYellowButtonMenu
	"Initialize the yellow button pop-up menu and corresponding messages."
	"SmalltalkEditor initialize"
	yellowButtonMenu := MenuMorph new  defaultTarget: self.
	yellowButtonMenu addList: {
		{'find...(f)' translated.				#find}.
		{'find again (g)' translated.			#findAgain}.
		{'set search string (h)' translated.	#setSearchString}.
		#-.	     
		{'do again (j)' translated.			#again}.
		{'undo (z)' translated.				#undo}.
		#-.	     
		{'copy (c)' translated.				#copySelection}.
		{'cut (x)' translated.					#cut}.
		{'paste (v)' translated.				#paste}.
		{'paste...' translated.					#pasteRecent}.
		#-.	     
		{'do it (d)' translated.				#doIt}.
		{'print it (p)' translated.				#printIt}.
		{'inspect it (i)' translated.			#inspectIt}.
		{'explore it (I)' translated.			#exploreIt}.
		{'debug it' translated.				#debugIt}.
		#-.	     
		{'accept (s)' translated.				#accept}.
		{'cancel (l)' translated.				#cancel}.
		#-.	     
		{'show bytecodes' translated.		#showBytecodes}.
		#-.	     
		{'more...' translated.					#shiftedTextPaneMenuRequest}.
	}! !

!SmalltalkEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:40'!
argAdvance: aKeyboardEvent
	"Invoked by Ctrl-a.  Useful after Ctrl-q.
	 Search forward from the end of the selection for a colon followed by
		a space.  Place the caret after the space.  If none are found, place the
		caret at the end of the text.  Does not affect the undoability of the 
	 	previous command."

	| start |
	self closeTypeIn.
	start := paragraph text findString: ': ' startingAt: self stopIndex.
	start = 0 ifTrue: [ start := paragraph text size + 1].
	self selectAt: start + 2.
	^true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'jmv 8/12/2009 22:42'!
browseChangeSetsWithSelector
	"Determine which, if any, change sets have at least one change for the selected selector, independent of class"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ morph flash].
	ChangeSorter browseChangeSetsWithSelector: aSelector! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'sd 6/3/2010 13:02'!
browseClassFromIt
	"Launch a hierarchy browser for the class indicated by the current selection.  If multiple classes matching the selection exist, let the user choose among them."
	
	| selectedClass |
	self lineSelectAndEmptyCheck: [^ self].
	selectedClass := self systemNavigation 
				classFromPattern: (self selection string copyWithoutAll: CharacterSet crlf)
				withCaption: 'choose a class to browse...'.
	selectedClass isNil ifTrue: [^ self internalCallToImplementorsOfIt ].
	self systemNavigation browseClass: selectedClass .! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'AlainPlantec 4/15/2010 11:38'!
browseIt
	"Launch a browser for the current selection, if appropriate"
	^ self browseClassFromIt
! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:40'!
browseIt: aKeyboardEvent
	"Triggered by Cmd-B; browse the thing represented by the current selection, if plausible.  1/18/96 sw"

	self browseIt.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'jmv 8/12/2009 22:53'!
browseItHere
	"Retarget the receiver's window to look at the selected class, if appropriate.  3/1/96 sw"
	| aSymbol foundClass b |
	(((b := model) isKindOf: Browser) and: [b couldBrowseAnyClass])
		ifFalse: [^ morph flash].
	model okToChange ifFalse: [^ morph flash].
	self selectionInterval isEmpty ifTrue: [self selectWord].
	(aSymbol := self selectedSymbol) isNil ifTrue: [^ morph flash].

	foundClass := (Smalltalk at: aSymbol ifAbsent: [nil]).
		foundClass isNil ifTrue: [^ morph flash].
		(foundClass isKindOf: Class)
			ifTrue:
				[model systemCategoryListIndex: 
					(model systemCategoryList indexOf: foundClass category).
	model classListIndex: (model classList indexOf: foundClass name)]! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:40'!
browseItHere: aKeyboardEvent
	"Triggered by Cmd-shift-B; browse the thing represented by the current selection, if plausible, in the receiver's own window.  3/1/96 sw"

	self browseItHere.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'FernandoOlivero 5/17/2010 13:59'!
classCommentsContainingIt
	"Open a browser class comments which contain the current selection somewhere in them."

	self lineSelectAndEmptyCheck: [^ self].
	self systemNavigation  browseClassCommentsWithString: self selection string! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'FernandoOlivero 5/17/2010 13:59'!
classNamesContainingIt
	"Open a browser on classes whose names contain the selected string"

	self lineSelectAndEmptyCheck: [^self].
	self systemNavigation 
		browseClassesWithNamesContaining: self selection string
		caseSensitive: Sensor leftShiftDown! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'FernandoOlivero 5/17/2010 14:05'!
compileSelectionFor: anObject in: evalContext
	| methodNode method failed |
	 failed :=  false.
	methodNode :=  
		[
			Compiler new
			compileNoPattern: self selectionAsStream
			in: anObject class
			context: evalContext
			notifying: self
			ifFail: [ failed :=  true ] ]
		on: OutOfScopeNotification
		do: [:ex | ex resume: true].
	failed ifTrue: [ ^ nil ].	
	self halt: 'old generate: use'.
	method := methodNode generate: #(0 0 0 0).
	^method copyWithTempsFromMethodNode: methodNode! !

!SmalltalkEditor methodsFor: 'private' stamp: 'StephaneDucasse 5/1/2010 16:13'!
completeSymbol: hintText lastOffering: selectorOrNil
	"Invoked by Ctrl-q when there is only a caret.
		Do selector-completion, i.e., try to replace the preceding identifier by a
		selector that begins with those characters & has as many keywords as possible.
	 	Leave two spaces after each colon (only one after the last) as space for
		arguments.  Put the caret after the space after the first keyword.  If the
		user types Ctrl-q again immediately, choose a different selector.
	 Undoer: #undoQuery:lastOffering:; Redoer: itself.
	If redoing, just redisplay the last offering, selector[OrNil]."

	| firstTime input prior caret newStart sym kwds outStream |
	firstTime := self isRedoing
		ifTrue: [prior := sym := selectorOrNil. true]
		ifFalse: [hintText isNil].
	firstTime
		ifTrue: "Initial Ctrl-q (or redo)"					
			[caret := self startIndex.
			self selectPrecedingIdentifier.
			input := self selection]
		ifFalse: "Repeated Ctrl-q"
			[caret := UndoInterval first + hintText size.
			self selectInvisiblyFrom: UndoInterval first to: UndoInterval last.
			input := hintText.
			prior := selectorOrNil].
	(input size ~= 0 and: [sym ~~ nil or:
			[(sym := Symbol thatStarts: input string skipping: prior) ~~ nil]])
		ifTrue: "found something to offer"
			[newStart := self startIndex.
			outStream := (String new: 2 * sym size) writeStream.
			1 to: (kwds := sym keywords) size do:
				[:i |
				outStream nextPutAll: (kwds at: i).
				i = 1 ifTrue: [caret := newStart + outStream contents size + 1].
				outStream nextPutAll:
					(i < kwds size ifTrue: ['  '] ifFalse: [' '])].
			UndoSelection := input.
			self deselect; zapSelectionWith: outStream contents asText.
			self undoer: #undoQuery:lastOffering: with: input with: sym]
		ifFalse: "no more matches"
			[firstTime ifFalse: "restore original text & set up for a redo"
				[UndoSelection := self selection.
				self deselect; zapSelectionWith: input.
				self undoer: #completeSymbol:lastOffering: with: input with: prior.
				Undone := true].
			morph flash].
	self selectAt: caret! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'jmv 8/12/2009 22:57'!
debug: aCompiledMethod receiver: anObject in: evalContext

	| selector guineaPig debugger context |
	selector := evalContext isNil ifTrue: [#DoIt] ifFalse: [#DoItIn:].
	anObject class addSelectorSilently: selector withMethod: aCompiledMethod.
	guineaPig := evalContext isNil
		ifTrue: [[anObject DoIt] newProcess]
		ifFalse: [[anObject DoItIn: evalContext] newProcess].
	context := guineaPig suspendedContext.
	debugger := Debugger new
		process: guineaPig
		controller: nil
		context: context.
	debugger openFullNoSuspendLabel: 'Debug it'.
	[debugger interruptedContext method == aCompiledMethod]
		whileFalse: [debugger send].
	anObject class basicRemoveSelector: selector! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'jmv 8/12/2009 22:31'!
debugIt

	| method receiver context |
	(model respondsTo: #doItReceiver) 
		ifTrue: 
			[FakeClassPool adopt: model selectedClass.
			receiver := model doItReceiver.
			context := model doItContext]
		ifFalse:
			[receiver := context := nil].
	self lineSelectAndEmptyCheck: [^self].
	method := self compileSelectionFor: receiver in: context.
	method notNil ifTrue:
		[self debug: method receiver: receiver in: context].
	FakeClassPool adopt: nil! !

!SmalltalkEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:40'!
displayIfFalse: aKeyboardEvent
	"Replace the current text selection with the text 'ifFalse:'--initiated by 
	ctrl-f."

	self addString: 'ifFalse:'.
	^false! !

!SmalltalkEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:40'!
displayIfTrue: aKeyboardEvent
	"Replace the current text selection with the text 'ifTrue:'--initiated by 
	ctrl-t."

	self addString: 'ifTrue:'.
	^false! !

!SmalltalkEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:41'!
doAgainOnce: aKeyboardEvent 
	"Do the previous thing again once. 1/26/96 sw"

	self closeTypeIn.
	self again.
	^ true! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'jmv 8/12/2009 22:34'!
doIt
	"Set the context to include pool vars of the model.  Then evaluate."
	^ self evaluateSelection! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
doIt: aKeyboardEvent
	"Called when user hits cmd-d.  Select the current line, if relevant, then evaluate and execute.  2/1/96 sw.
	2/29/96 sw: don't call selectLine; it's done by doIt now"

	self doIt.
	^ true! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'StephaneDucasse 7/10/2010 17:23'!
evaluateSelection
	"Treat the current selection as an expression; evaluate it and return the result"
	| result rcvr ctxt |
	self lineSelectAndEmptyCheck: [^ ''].
	self flag: #todo." NoDoitReceiverNormally now"
	(model respondsTo: #doItReceiver) 
		ifTrue: [FakeClassPool adopt: model selectedClass.  "Include model pool vars if any"
				rcvr := model doItReceiver.
				ctxt := model doItContext]
		ifFalse: [rcvr := ctxt := nil].
	result := [
		rcvr class evaluatorClass new 
			evaluate: self selectionForDoitAsStream
			in: ctxt
			to: rcvr
			notifying: self
			ifFail: [FakeClassPool adopt: nil. ^ #failedDoit]
			logged: true.
	] 
		on: OutOfScopeNotification 
		do: [ :ex | ex resume: true].
	FakeClassPool adopt: nil.
	^ result! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'jmv 8/12/2009 22:55'!
explain
	"Try to shed some light on what kind of entity the current selection is. 
	The selection must be a single token or construct. Insert the answer after 
	the selection. Send private messages whose names begin with 'explain' 
	that return a string if they recognize the selection, else nil."

	| string tiVars cgVars selectors delimitors numbers sorry reply symbol |
Cursor execute showWhile: 
			[sorry := '"Sorry, I can''t explain that.  Please select a single
token, construct, or special character.'.
			sorry := sorry , (morph canDiscardEdits
							ifFalse: ['  Also, please cancel or accept."']
							ifTrue: ['"']).
			(string := self selection asString) isEmpty
				ifTrue: [reply := '']
				ifFalse: [string := self explainScan: string.
					"Remove space, tab, cr"
					"Temps and Instance vars need only test strings that are all letters"
					(string detect: [:char | (char isLetter or: [char isDigit]) not]
						ifNone: []) ifNil: 
							[tiVars := self explainTemp: string.
							tiVars == nil ifTrue: [tiVars := self explainInst: string]].
					(tiVars == nil and: [model respondsTo: #explainSpecial:])
						ifTrue: [tiVars := model explainSpecial: string].
					tiVars == nil
						ifTrue: [tiVars := '']
						ifFalse: [tiVars := tiVars , '\' withCRs].
					"Context, Class, Pool, and Global vars, and Selectors need 
					only test symbols"
					(Symbol hasInterned: string ifTrue: [:s | symbol := s])
						ifTrue: [cgVars := self explainCtxt: symbol.
							cgVars == nil
								ifTrue: [cgVars := self explainClass: symbol.
									cgVars == nil ifTrue: [cgVars := self explainGlobal: symbol]].
							"See if it is a Selector (sent here or not)"
							selectors := self explainMySel: symbol.
							selectors == nil
								ifTrue: 
									[selectors := self explainPartSel: string.
									selectors == nil ifTrue: [
										selectors := self explainAnySel: symbol]]]
						ifFalse: [selectors := self explainPartSel: string].
					cgVars == nil
						ifTrue: [cgVars := '']
						ifFalse: [cgVars := cgVars , '\' withCRs].
					selectors == nil
						ifTrue: [selectors := '']
						ifFalse: [selectors := selectors , '\' withCRs].
					string size = 1
						ifTrue: ["single special characters"
							delimitors := self explainChar: string]
						ifFalse: ["matched delimitors"
							delimitors := self explainDelimitor: string].
					numbers := self explainNumber: string.
					numbers == nil ifTrue: [numbers := ''].
					delimitors == nil ifTrue: [delimitors := ''].
					reply := tiVars , cgVars , selectors , delimitors , numbers].
			reply size = 0 ifTrue: [reply := sorry].
			self afterSelectionInsertAndSelect: reply]! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'FernandoOlivero 5/17/2010 14:06'!
explainAnySel: symbol 
	"Is this any message selector?"

	| list reply |
	list := self systemNavigation  allClassesImplementing: symbol.
	list size = 0 ifTrue: [^nil].
	list size < 12
		ifTrue: [reply := ' is a message selector which is defined in these classes ' , list printString]
		ifFalse: [reply := ' is a message selector which is defined in many classes'].
	^'"' , symbol , reply , '."' , '\' withCRs, ' SystemNavigation default  browseAllImplementorsOf: #' , symbol! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'FernandoOlivero 3/12/2010 23:57'!
explainChar: string
	"Does string start with a special character?"

	| char |
	char := string at: 1.
	char = $. ifTrue: [^'"Period marks the end of a Smalltalk statement.  A period in the middle of a number means a decimal point.  (The number is an instance of class Float)."'].
	char = $' ifTrue: [^'"The characters between two single quotes are made into an instance of class String"'].
	char = $" ifTrue: [^'"Double quotes enclose a comment.  Smalltalk ignores everything between double quotes."'].
	char = $# ifTrue: [^'"The characters following a hash mark are made into an instance of class Symbol.  If parenthesis follow a hash mark, an instance of class Array is made.  It contains literal constants."'].
	(char = $( or: [char = $)]) ifTrue: [^'"Expressions enclosed in parenthesis are evaluated first"'].
	(char = $[ or: [char = $]]) ifTrue: [^'"The code inside square brackets is an unevaluated block of code.  It becomes an instance of BlockContext and is usually passed as an argument."'].
	(char = ${ or: [char = $}]) ifTrue: [^ '"A sequence of expressions separated by periods, when enclosed in curly braces, are evaluated to yield the elements of a new Array"'].
	(char = $< or: [char = $>]) ifTrue: [^'"<primitive: xx> means that this method is usually preformed directly by the virtual machine.  If this method is primitive, its Smalltalk code is executed only when the primitive fails."'].
	char = $^ ifTrue: [^'"Uparrow means return from this method.  The value returned is the expression following the ^"'].
	char = $| ifTrue: [^'"Vertical bars enclose the names of the temporary variables used in this method.  In a block, the vertical bar separates the argument names from the rest of the code."'].
	char = $; ifTrue: [^'"Semicolon means cascading.  The message after the semicolon is sent to the same object which received the message before the semicolon."'].
	char = $: ifTrue: [^'"A colon at the end of a keyword means that an argument is expected to follow.  Methods which take more than one argument have selectors with more than one keyword.  (One keyword, ending with a colon, appears before each argument).', '\\' withCRs, 'A colon before a variable name just inside a block means that the block takes an agrument.  (When the block is evaluated, the argument will be assigned to the variable whose name appears after the colon)."'].
	char = $$ ifTrue: [^'"The single character following a dollar sign is made into an instance of class Character"'].
	char = $- ifTrue: [^'"A minus sign in front of a number means a negative number."'].
	char = $e ifTrue: [^'"An e in the middle of a number means that the exponent follows."'].
	char = $r ifTrue: [^'"An r in the middle of a bunch of digits is an instance of Integer expressed in a certain radix.  The digits before the r denote the base and the digits after it express a number in that base."'].
	char = Character space ifTrue: [^'"the space Character"'].
	char = Character tab ifTrue: [^'"the tab Character"'].
	char = Character cr ifTrue: [^'"the carriage return Character"'].
	^nil! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'StephaneDucasse 5/1/2010 16:13'!
explainClass: symbol 
	"Is symbol a class variable or a pool variable?"
	| class reply classes |
	(model respondsTo: #selectedClassOrMetaClass)
		ifFalse: [^ nil].
	(class := model selectedClassOrMetaClass) ifNil: [^ nil].
	"no class is selected"
	(class isKindOf: Metaclass)
		ifTrue: [class := class soleInstance].
	classes := (Array with: class)
				, class allSuperclasses.
	"class variables"
	reply := classes detect: [:each | (each classVarNames detect: [:name | symbol = name]
					ifNone: [])
					~~ nil]
				ifNone: [].
	reply == nil ifFalse: [^ '"is a class variable, defined in class ' , reply printString , '"\' withCRs , 'SystemNavigation new browseAllCallsOn: (' , reply printString , ' classPool associationAt: #' , symbol , ').'].
	"pool variables"
	classes do: [:each | (each sharedPools
			detect: [:pool | (pool includesKey: symbol)
					and: 
						[reply := pool.
						true]]
			ifNone: [])
			~~ nil].
	reply
		ifNil: [(Undeclared includesKey: symbol)
				ifTrue: [^ '"is an undeclared variable.' , '"\' withCRs , 'SystemNavigation new browseAllCallsOn: (Undeclared associationAt: #' , symbol , ').']]
		ifNotNil: 
			[classes := Array new writeStream.
			Smalltalk
				allBehaviorsDo: [:each | (each sharedPools
						detect: 
							[:pool | 
							pool == reply]
						ifNone: [])
						~~ nil ifTrue: [classes nextPut: each]].
			"Perhaps not print whole list of classes if too long. (unlikely)"
			^ '"is a pool variable from the pool ' , (Smalltalk keyAtIdentityValue: reply) asString , ', which is used by the following classes ' , classes contents printString , '"\' withCRs , 'SystemNavigation new browseAllCallsOn: (' , (Smalltalk keyAtIdentityValue: reply) asString , ' bindingOf: #' , symbol , ').'].
	^ nil! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'jmv 8/12/2009 23:00'!
explainCtxt: symbol 
	"Is symbol a context variable?"

	| reply classes text cls |
	symbol = #nil ifTrue: [reply := '"is a constant.  It is the only instance of class UndefinedObject.  nil is the initial value of all variables."'].
	symbol = #true ifTrue: [reply := '"is a constant.  It is the only instance of class True and is the receiver of many control messages."'].
	symbol = #false ifTrue: [reply := '"is a constant.  It is the only instance of class False and is the receiver of many control messages."'].
	symbol = #thisContext ifTrue: [reply := '"is a context variable.  Its value is always the MethodContext which is executing this method."'].
	(model respondsTo: #selectedClassOrMetaClass) ifTrue: [
		cls := model selectedClassOrMetaClass].
	cls ifNil: [^ reply].	  "no class known"
	symbol = #self ifTrue: 
			[classes := cls withAllSubclasses.
			classes size > 12
				ifTrue: [text := cls printString , ' or a subclass']
				ifFalse: 
					[classes := classes printString.
					text := 'one of these classes' , (classes copyFrom: 4 to: classes size)].
			reply := '"is the receiver of this message; an instance of ' , text , '"'].
	symbol = #super ifTrue: [reply := '"is just like self.  Messages to super are looked up in the superclass (' , cls superclass printString , ')"'].
	^reply! !

!SmalltalkEditor methodsFor: 'private' stamp: 'jmv 12/31/2009 10:10'!
explainDelimitor: string
	"Is string enclosed in delimitors?"

	| str |
	(string at: 1) isLetter ifTrue: [^nil].  "only special chars"
	(string first = string last) ifTrue:
			[^ self explainChar: (String with: string first)]
		ifFalse:
			[(string first = $( and: [string last = $)]) ifTrue:
				[^ self explainChar: (String with: string first)].
			(string first = $[ and: [string last = $]]) ifTrue:
				[^ self explainChar: (String with: string first)].
			(string first = ${ and: [string last = $}]) ifTrue:
				[^ self explainChar: (String with: string first)].
			(string first = $< and: [string last = $>]) ifTrue:
				[^ self explainChar: (String with: string first)].
			(string first = $# and: [string last = $)]) ifTrue:
				[^'"An instance of class Array.  The Numbers, Characters, or Symbols between the parenthesis are the elements of the Array."'].
			string first = $# ifTrue:
				[^'"An instance of class Symbol."'].
			(string first = $$ and: [string size = 2]) ifTrue:
				[^'"An instance of class Character.  This one is the character ', (String with: string last), '."'].
			(string first = $:) ifTrue:
				[str := string allButFirst.
				(self explainTemp: str) ~~ nil ifTrue:
					[^'"An argument to this block will be bound to the temporary variable ',
						str, '."']]].
	^ nil! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'simon.denier 6/11/2010 14:49'!
explainGlobal: symbol 
	"Is symbol a global variable?"
	| reply classes |
	reply := Smalltalk globals at: symbol ifAbsent: [^nil].
	(reply class == Dictionary or: [reply isKindOf: SharedPool class])
		ifTrue: 
			[classes := Set new.
			Smalltalk allBehaviorsDo: [:each | (each sharedPools detect: [:pool | pool == reply]
					ifNone: [])
					~~ nil ifTrue: [classes add: each]].
			classes := classes printString.
			^'"is a global variable.  It is a pool which is used by the following classes ' , (classes allButFirst: 5) , '"'].
	(reply isKindOf: Behavior)
		ifTrue: [^'"is a global variable.  ' , symbol , ' is a class in category ', reply category,
			'."', '\' withCRs, 'Browser newOnClass: ' , symbol , '.'].
	symbol == #Smalltalk ifTrue: [^'"is a global.  Smalltalk is the only instance of SystemDictionary and holds all global variables."'].
	^'"is a global variable.  ' , symbol , ' is ' , reply printString , '"'! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'jmv 8/12/2009 23:01'!
explainInst: string 
	"Is string an instance variable of this class?"
	| classes cls |

	(model respondsTo: #selectedClassOrMetaClass) ifTrue: [
		cls := model selectedClassOrMetaClass].
	cls ifNil: [^ nil].	  "no class known"
	classes := (Array with: cls)
				, cls allSuperclasses.
	classes := classes detect: [:each | (each instVarNames
			detect: [:name | name = string] ifNone: [])
			~~ nil] ifNone: [^nil].
	classes := classes printString.
	^ '"is an instance variable of the receiver; defined in class ' , classes , 
		'"\' withCRs , classes , ' systemNavigation browseAllAccessesTo: ''' , string , ''' from: ', classes, '.'! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'jmv 8/12/2009 23:01'!
explainMySel: symbol 
	"Is symbol the selector of this method?  Is it sent by this method?  If 
	not, then expalin will call (explainPartSel:) to see if it is a fragment of a 
	selector sent here.  If not, explain will call (explainAnySel:) to catch any 
	selector. "

	| lits classes msg |
	(model respondsTo: #selectedMessageName) ifFalse: [^ nil].
	(msg := model selectedMessageName) ifNil: [^nil].	"not in a message"
	classes := Smalltalk allClassesImplementing: symbol.
	classes size > 12
		ifTrue: [classes := 'many classes']
		ifFalse: [classes := 'these classes ' , classes printString].
	msg = symbol
		ifTrue: [^ '"' , symbol , ' is the selector of this very method!!  It is defined in ',
			classes , '.  To see the other definitions, go to the message list pane, get the menu from the top of the scroll bar, and select ''implementors of...''."']
		ifFalse: 
			[lits := (model selectedClassOrMetaClass compiledMethodAt:
				msg) messages.
			(lits detect: [:each | each == symbol]
				ifNone: [])
				== nil ifTrue: [^nil].
			^ '"' , symbol , ' is a message selector which is defined in ', classes , '.  To see the definitions, go to the message list pane, get the menu from the top of the scroll bar, and select ''implementors of...''."'].! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'jmv 8/12/2009 23:01'!
explainNumber: string 
	"Is string a Number?"

	| strm c |
	(c := string at: 1) isDigit ifFalse: [(c = $- and: [string size > 1 and: [(string at: 2) isDigit]])
			ifFalse: [^nil]].
	strm := ReadStream on: string.
	c := Number readFrom: strm.
	strm atEnd ifFalse: [^nil].
	c printString = string
		ifTrue: [^'"' , string , ' is a ' , c class name , '"']
		ifFalse: [^'"' , string , ' (= ' , c printString , ') is a ' , c class name , '"']! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'jmv 8/12/2009 22:59'!
explainPartSel: string 
	"Is this a fragment of a multiple-argument selector sent in this method?"
	| lits whole reply classes s msg |

	(model respondsTo: #selectedMessageName) ifFalse: [^ nil].
	(msg := model selectedMessageName) ifNil: [^ nil].  "not in a message"
	string last == $: ifFalse: [^ nil].
	"Name of this method"
	lits := Array with: msg.
	(whole := lits detect: [:each | (each keywords detect: [:frag | frag = string]
					ifNone: []) ~~ nil]
				ifNone: []) ~~ nil
		ifTrue: [reply := ', which is the selector of this very method!!'.
			s := '.  To see the other definitions, go to the message list pane, get the menu from the top of the scroll bar, and select ''implementors of...''."']
		ifFalse: 
			["Selectors called from this method"
			lits := (model selectedClassOrMetaClass compiledMethodAt:
				msg) messages.
			(whole := lits detect: [:each | (each keywords detect: [:frag | frag = string]
							ifNone: []) ~~ nil]
						ifNone: []) ~~ nil
				ifFalse: [string = 'primitive:'
					ifTrue: [^self explainChar: '<']
					ifFalse: [^nil]].
			reply := '.'.
			s := '.  To see the definitions, go to the message list pane, get the menu from the top of the scroll bar, and select ''implementors of...''."'].
	classes := Smalltalk allClassesImplementing: whole.
	classes size > 12
		ifTrue: [classes := 'many classes']
		ifFalse: [classes := 'these classes ' , classes printString].
	^ '"' , string , ' is one part of the message selector ' , whole, reply , '  It is defined in ' , classes , s! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'jmv 8/12/2009 23:01'!
explainScan: string 
	"Remove beginning and trailing space, tab, cr.
	 1/15/96 sw: copied intact from BrowserCodeController"

	| c beg end |
	beg := 1.
	end := string size.
	
	[beg = end ifTrue: [^string copyFrom: 1 to: 1].
	"if all blank, tell about the first"
	c := string at: beg.
	c = Character space or: [c = Character tab or: [c = Character cr]]]
		whileTrue: [beg := beg + 1].
	
	[c := string at: end.
	c = Character space or: [c = Character tab or: [c = Character cr]]]
		whileTrue: [end := end - 1].
	^string copyFrom: beg to: end	"Return purely visible characters"! !

!SmalltalkEditor methodsFor: 'explain' stamp: 'jmv 8/12/2009 23:01'!
explainTemp: string 
	"Is string the name of a temporary variable (or block argument variable)?"

	| selectedClass tempNames i reply methodNode method msg |
	(model respondsTo: #selectedMessageName) ifFalse: [^ nil].
	(msg := model selectedMessageName) ifNil: [^nil].	"not in a message"
	selectedClass := model selectedClassOrMetaClass.
	tempNames := selectedClass parserClass new 
			parseArgsAndTemps: model selectedMessage notifying: nil.
	method := selectedClass compiledMethodAt: msg.
	(i := tempNames findFirst: [:each | each = string]) = 0 ifTrue: [
		(method numTemps > tempNames size)
			ifTrue: 
				["It must be an undeclared block argument temporary"
				methodNode := selectedClass compilerClass new
							parse: model selectedMessage
							in: selectedClass
							notifying: nil.
				tempNames := methodNode tempNames]
			ifFalse: [^nil]].
	(i := tempNames findFirst: [:each | each = string]) > 0 ifTrue: [i > method numArgs
			ifTrue: [reply := '"is a temporary variable in this method"']
			ifFalse: [reply := '"is an argument to this method"']].
	^reply! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'jmv 8/12/2009 22:39'!
exploreIt
	| result |
	result := self evaluateSelection.
	((result isKindOf: FakeClassPool) or: [result == #failedDoit])
			ifTrue: [morph flash]
			ifFalse: [result explore]! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
exploreIt: aKeyboardEvent
	"Explore the selection -- invoked via cmd-shift-I.  If there is no current selection, use the current line."

	self exploreIt.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'jmv 8/12/2009 22:44'!
fileItIn
	"Make a Stream on the text selection and fileIn it.
	 1/24/96 sw: moved here from FileController; this function can be useful from any text window that shows stuff in chunk format"

	| selection |
	selection := self selection.
	(ReadWriteStream on: selection string from: 1 to: selection size) fileIn
! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
fileItIn: aKeyboardEvent
	"File in the selection; invoked via a keyboard shortcut, -- for now, cmd-shift-G."

	self fileItIn.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'sd 6/3/2010 13:03'!
implementorsOfIt
	"Open an implementors browser on the selected selector"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ morph flash].
	(self systemNavigation allImplementorsOf: aSelector) isEmpty
			ifTrue: [self internalCallToBrowseIt]
			ifFalse: [ self systemNavigation browseAllImplementorsOf: aSelector]! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
implementorsOfIt: aKeyboardEvent
	"Triggered by Cmd-m; browse implementors of the selector represented by the current selection, if plausible. 2/1/96 sw"

	self implementorsOfIt.
	^ true! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'jmv 8/12/2009 22:39'!
inspectIt
	| result |
	result := self evaluateSelection.
	((result isKindOf: FakeClassPool) or: [result == #failedDoit])
			ifTrue: [morph flash]
			ifFalse: [result inspect]! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
inspectIt: aKeyboardEvent
	"Inspect the selection -- invoked via cmd-i.  If there is no current selection, use the current line.  1/17/96 sw
	 2/29/96 sw: don't call selectLine; it's done by inspectIt now"

	self inspectIt.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'sd 6/3/2010 13:06'!
internalCallToBrowseIt
	"Launch a browser for the class indicated by the current selection. 
	If multiple classes matching the selection exist, let the user choose among them."
	| aClass |
	self lineSelectAndEmptyCheck: [^ self].
	aClass := self systemNavigation 
				classFromPattern: (self selection string copyWithoutAll: CharacterSet crlf)
				withCaption: 'choose a class to browse...'.
	aClass ifNil: [^ self].
	self systemNavigation browseClass: aClass .! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'sd 6/3/2010 13:05'!
internalCallToImplementorsOfIt
	"Open an implementors browser on the selected selector"

	| aSelector |
	self lineSelectAndEmptyCheck: [^ self].
	(aSelector := self selectedSelector) == nil ifTrue: [^ morph flash].
	 (self systemNavigation allImplementorsOf: aSelector) isEmpty
				ifFalse: [ self systemNavigation browseAllImplementorsOf: aSelector]
			
	! !

!SmalltalkEditor methodsFor: 'testing' stamp: 'FernandoOlivero 6/15/2010 13:07'!
isSmalltalkEditor
	^ true ! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'FernandoOlivero 5/17/2010 14:07'!
methodNamesContainingIt
	"Open a browser on methods names containing the selected string"

	self lineSelectAndEmptyCheck: [^ self].
	Cursor wait showWhile: [ self systemNavigation  browseMethodsWhoseNamesContain: self selection string withBlanksTrimmed].
	! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
methodNamesContainingIt: aKeyboardEvent
	"Browse methods whose selectors containing the selection in their names"

	self methodNamesContainingIt.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'FernandoOlivero 5/17/2010 14:08'!
methodSourceContainingIt
	"Open a browser on methods which contain the current selection in their source (case-sensitive full-text search of source).   EXTREMELY slow!!"
	self lineSelectAndEmptyCheck: [^ self].
	(self confirm: 'This will take a few minutes. Shall I proceed?') ifFalse: [^ self].
	self systemNavigation  browseMethodsWithSourceString: self selection string! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
methodStringsContainingIt: aKeyboardEvent
	"Invoked from cmd-E -- open a browser on all methods holding string constants containing it."

	self methodStringsContainingit.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'FernandoOlivero 5/17/2010 14:08'!
methodStringsContainingit
	"Open a browser on methods which contain the current selection as part of a string constant."

	self lineSelectAndEmptyCheck: [^ self].
	self systemNavigation  browseMethodsWithString: self selection string! !

!SmalltalkEditor methodsFor: 'new selection' stamp: 'jmv 8/12/2009 23:04'!
nextTokenFrom: start direction: dir
	"simple token-finder for compiler automated corrections"
	| loc str |
	loc := start + dir.
	str := paragraph text string.
	[(loc between: 1 and: str size) and: [(str at: loc) isSeparator]]
		whileTrue: [loc := loc + dir].
	^ loc! !

!SmalltalkEditor methodsFor: 'new selection' stamp: 'jmv 8/12/2009 23:04'!
notify: aString at: anInteger in: aStream 
	"The compilation of text failed. The syntax error is noted as the argument, 
	aString. Insert it in the text at starting character position anInteger."

	| pos |
	pos := self selectionInterval notEmpty
		ifTrue: [
			self startIndex + anInteger - 1 ]
		ifFalse: [anInteger].
	self insertAndSelect: aString at: (pos max: 1)! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
pasteInitials: aKeyboardEvent 
	"Replace the current text selection by an authorship name/date stamp; invoked by cmd-shift-v, easy way to put an authorship stamp in the comments of an editor."

	self closeTypeIn.
	self replace: self selectionInterval with: (Text fromString: Utilities changeStamp) and: [self selectAt: self stopIndex].
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'jmv 8/18/2009 18:00'!
prettyPrint
	"Reformat the contents of the receiver's view (a Browser)."

	| selectedClass newText |
	model selectedMessageName ifNil: [^ morph flash].
	selectedClass := model selectedClassOrMetaClass.
	newText := selectedClass compilerClass new
		format: self text
		in: selectedClass
		notifying: self.
	newText ifNotNil:
		[self deselect; selectInvisiblyFrom: 1 to: paragraph text size.
		self replaceSelectionWith: (newText asText makeSelectorBoldIn: selectedClass).
		self selectAt: 1]! !

!SmalltalkEditor methodsFor: 'do-its' stamp: 'jmv 8/12/2009 22:39'!
printIt
	"Treat the current text selection as an expression; evaluate it. Insert the 
	description of the result of evaluation after the selection and then make 
	this description the new text selection."
	| result |
	result := self evaluateSelection.
	((result isKindOf: FakeClassPool) or: [result == #failedDoit])
			ifTrue: [morph flash]
			ifFalse: [self afterSelectionInsertAndSelect: result printString]! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
printIt: aKeyboardEvent
	"Print the results of evaluting the selection -- invoked via cmd-p.  If there is no current selection, use the current line.  1/17/96 sw
	 2/29/96 sw: don't call selectLine now, since it's called by doIt"

	self printIt.
	^ true! !

!SmalltalkEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:41'!
querySymbol: aKeyboardEvent
	"Invoked by Ctrl-q to query the Symbol table and display alternate symbols.
	 See comment in completeSymbol:lastOffering: for details."

	self closeTypeIn.
	self hasCaret
		ifTrue: "Ctrl-q typed when a caret"
			[self perform: #completeSymbol:lastOffering: withArguments:
				((UndoParagraph == paragraph and: [UndoMessage sends: #undoQuery:lastOffering:])
					ifTrue: [UndoMessage arguments] "repeated Ctrl-q"
					ifFalse: [Array with: nil with: nil])] "initial Ctrl-q"
		ifFalse: "Ctrl-q typed when statements were highlighted"
			[morph flash].
	^true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'FernandoOlivero 5/17/2010 14:08'!
referencesToIt
	"Open a references browser on the selected symbol"

	| aSymbol |
	self selectLine.
	((aSymbol := self selectedSymbol) == nil or:
		[(Smalltalk includesKey: aSymbol) not])
			ifTrue: [^ morph flash].

	self systemNavigation  browseAllCallsOn: (Smalltalk associationAt: self selectedSymbol)! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
referencesToIt: aKeyboardEvent
	"Triggered by Cmd-N; browse references to the current selection"

	self referencesToIt.
	^ true! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
save: aKeyboardEvent
	"Submit the current text.  Equivalent to 'accept' 1/18/96 sw"

	self closeTypeIn.
	self accept.
	^ true! !

!SmalltalkEditor methodsFor: 'new selection' stamp: 'jmv 8/12/2009 23:06'!
selectPrecedingIdentifier
	"Invisibly select the identifier that ends at the end of the selection, if any."

	| string sep stop tok |
	tok := false.
	string := paragraph text string.
	stop := self stopIndex - 1.
	[stop > 0 and: [(string at: stop) isSeparator]] whileTrue: [stop := stop - 1].
	sep := stop.
	[sep > 0 and: [(string at: sep) tokenish]] whileTrue: [tok := true. sep := sep - 1].
	tok ifTrue: [self selectInvisiblyFrom: sep + 1 to: stop]! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'jmv 8/12/2009 23:03'!
selectedSelector
	"Try to make a selector out of the current text selection"
	^self selection string findSelector! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'jmv 8/12/2009 23:03'!
selectedSymbol
	"Return the currently selected symbol, or nil if none.  Spaces, tabs and returns are ignored"

	| aString |
	self hasCaret ifTrue: [^ nil].
	aString := self selection string copyWithoutAll:
		{Character space.  Character cr.  Character tab}.
	aString size = 0 ifTrue: [^ nil].
	Symbol hasInterned: aString  ifTrue: [:sym | ^ sym].

	^ nil! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'FernandoOlivero 5/17/2010 14:10'!
sendersOfIt
	"Open a senders browser on the selected selector"

	| selectedSelector |
	self lineSelectAndEmptyCheck: [^ self].
	selectedSelector  := self selectedSelector.
	selectedSelector isNil  ifTrue: [^ morph flash].
	self systemNavigation  browseAllCallsOn: selectedSelector! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
sendersOfIt: aKeyboardEvent
	"Triggered by Cmd-n; browse implementors of the selector represented by the current selection, if plausible. 2/1/96 sw"

	self sendersOfIt.
	^ true! !

!SmalltalkEditor methodsFor: 'menu messages' stamp: 'jmv 8/12/2009 22:44'!
spawn
	"Create and schedule a message browser for the code of the model's 
	selected message. Retain any edits that have not yet been accepted."
	| code |
	code := paragraph text string.
	self cancel.
	model spawn: code.! !

!SmalltalkEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:41'!
spawnIt: aKeyboardEvent
	"Triggered by Cmd-o; spawn a new code window, if it makes sense."

	self spawn.
	^ true! !

!SmalltalkEditor methodsFor: 'undoers' stamp: 'jmv 8/12/2009 23:05'!
undoQuery: hintText lastOffering: selectorOrNil
	"Undo ctrl-q.  selectorOrNil (if not nil) is the previously offered selector.
	 hintText is the original hint.  Redoer: completeSymbol."

	self zapSelectionWith: UndoSelection.
	self undoMessage: (Message selector: #completeSymbol:lastOffering: arguments: UndoMessage arguments) forRedo: true.
	self selectAt: self stopIndex! !

!TextEditor class methodsFor: 'class initialization' stamp: 'jmv 11/5/2008 22:21'!
abandonChangeText
	"Call this to get out of the maddening situation in which the system keeps aggressively trying to do a replacement that you no longer wish to make, every time you make choose a new method in a list."
	ChangeText := FindText

	"
	TextEditor abandonChangeText
	"! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 11:34'!
caseSensitiveFinds
	^ CaseSensitiveFinds ifNil: [CaseSensitiveFinds := false]! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 11:34'!
caseSensitiveFinds: aBoolean
	CaseSensitiveFinds := aBoolean! !

!TextEditor class methodsFor: 'accessing' stamp: 'jmv 7/21/2009 10:07'!
cmdActions
	^cmdActions! !

!TextEditor class methodsFor: 'class initialization' stamp: 'FernandoOlivero 3/14/2010 23:18'!
initialize 
	"Initialize the keyboard shortcut maps and the shared buffers
	for copying text across views and managing again and undo." 
 
	"
	TextEditor initialize
	"

	UndoSelection := FindText := ChangeText := Text new.
	UndoMessage := Message selector: #halt.

	self initializeCmdKeyShortcuts.
	self initializeShiftCmdKeyShortcuts.
	self initializeYellowButtonMenu.
	self initializeShiftedYellowButtonMenu! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'FernandoOlivero 6/6/2010 15:56'!
initializeBasicCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table."

	"
	TextEditor initialize
	"

	| cmdMap cmds |
	cmdMap := Array new: 256 withAll: #noop:.		"use temp in case of a crash"
	cmdMap at: 1 + 1 put: #cursorHome:.				"home key"
	cmdMap at: 4 + 1 put: #cursorEnd:.				"end key"
	cmdMap at: 8 + 1 put: #backspace:.				"ctrl-H or delete key"
	cmdMap at: 11 + 1 put: #cursorPageUp:.			"page up key"
	cmdMap at: 12 + 1 put: #cursorPageDown:.		"page down key"
	cmdMap at: 13 + 1 put: #crWithIndent:.			"cmd-Return"
	cmdMap at: 27 + 1 put: #offerMenuFromEsc:.		"escape key"
	cmdMap at: 28 + 1 put: #cursorLeft:.				"left arrow key"
	cmdMap at: 29 + 1 put: #cursorRight:.			"right arrow key"
	cmdMap at: 30 + 1 put: #cursorUp:.				"up arrow key"
	cmdMap at: 31 + 1 put: #cursorDown:.			"down arrow key"
	cmdMap at: 32 + 1 put: #selectWord:.			"space bar key"
	cmdMap at: 127 + 1 put: #forwardDelete:.		"del key"
	cmdMap at: 115 + 1 put:#accept:.	
	'([{''"<' do: [:char | cmdMap at: char asciiValue + 1 put: #enclose:].
	
	cmdMap at: $, asciiValue + 1 put: #shiftEnclose:.
	
	cmds := #($a #selectAll: $c #copySelection: $e #exchange: $f #find: $g #findAgain: $h #setSearchString: $v #paste: $w #backWord: $x #cut: $y #swapChars: $z #undo:).
	1 to: cmds size
		by: 2
		do: [:i | cmdMap at: (cmds at: i) asciiValue + 1 put: (cmds at: i + 1)].
		
	cmdActions := cmdMap! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'FernandoOlivero 6/6/2010 15:51'!
initializeBasicShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."

	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	"TextEditor initialize"
	
	| cmdMap cmds |

	"shift-command and control shortcuts"
	cmdMap := Array new: 256 withAll: #noop:.  		"use temp in case of a crash"
	cmdMap at: ( 1 + 1) put: #cursorHome:.			"home key"
	cmdMap at: ( 4 + 1) put: #cursorEnd:.				"end key"
	cmdMap at: ( 8 + 1) put: #forwardDelete:.			"ctrl-H or delete key"
	cmdMap at: (11 + 1) put: #cursorPageUp:.			"page up key"
	cmdMap at: (12 + 1) put: #cursorPageDown:.		"page down key"
	cmdMap at: (13 + 1) put: #crWithIndent:.			"ctrl-Return"
	cmdMap at: (27 + 1) put: #offerMenuFromEsc:.	"escape key"
	cmdMap at: (28 + 1) put: #cursorLeft:.			"left arrow key"
	cmdMap at: (29 + 1) put: #cursorRight:.			"right arrow key"
	cmdMap at: (30 + 1) put: #cursorUp:.				"up arrow key"
	cmdMap at: (31 + 1) put: #cursorDown:.			"down arrow key"
	cmdMap at: (32 + 1) put: #selectWord:.			"space bar key"
	cmdMap at: (45 + 1) put: #changeEmphasis:.		"cmd-sh-minus"
	cmdMap at: (61 + 1) put: #changeEmphasis:.		"cmd-sh-plus"
	cmdMap at: (127 + 1) put: #forwardDelete:.		"del key"

	"Note: Command key overrides shift key, so, for example, cmd-shift-9 produces $9 not $("
	'9[,''' do: [ :char | cmdMap at: (char asciiValue + 1) put: #shiftEnclose: ].	"({< and double-quote"
	"Note: Must use cmd-9 or ctrl-9 to get '()' since cmd-shift-9 is a Mac FKey command."

	"NB: sw 12/9/2001 commented out the idiosyncratic line just below, which was grabbing shift-esc in the text editor and hence which argued with the wish to have shift-esc be a universal gesture for escaping the local context and calling up the desktop menu."  
	"cmdMap at: (27 + 1) put: #shiftEnclose:." 	"ctrl-["

	"'""''(' do: [ :char | cmdMap at: (char asciiValue + 1) put: #enclose:]."

	cmds := #(
		$c	compareToClipboard:
		$d	duplicate:
		$h	cursorTopHome:
		$j	doAgainMany:
		$k	changeStyle:
		$l	outdent:
		$m	selectCurrentTypeIn:
		$r	indent:
		$s	search:
		$u	changeLfToCr:
		$x	makeLowercase:
		$y	makeUppercase:
		$z	makeCapitalized:
	).
	1 to: cmds size by: 2 do: [ :i |
		cmdMap at: ((cmds at: i) asciiValue + 1) put: (cmds at: i + 1).			"plain keys"
		cmdMap at: ((cmds at: i) asciiValue - 32 + 1) put: (cmds at: i + 1).		"shifted keys"
		cmdMap at: ((cmds at: i) asciiValue - 96 + 1) put: (cmds at: i + 1).		"ctrl keys"
	].
	shiftCmdActions := cmdMap! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'FernandoOlivero 6/6/2010 15:51'!
initializeCmdKeyShortcuts
	"Initialize the (unshifted) command-key (or alt-key) shortcut table."



	"
	TextEditor initialize
	"

	| cmds |
	self initializeBasicCmdKeyShortcuts.
			
	'0123456789-=' 
		do: [:char | cmdActions at: char asciiValue + 1 put: #changeEmphasis:].
	
	cmds := #($k #offerFontMenu: $u #align: ).
	1 to: cmds size
		by: 2
		do: [:i | cmdActions at: (cmds at: i) asciiValue + 1 put: (cmds at: i + 1)]! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'FernandoOlivero 6/6/2010 15:51'!
initializeShiftCmdKeyShortcuts 
	"Initialize the shift-command-key (or control-key) shortcut table."
	"wod 11/3/1998: Fix setting of cmdMap for shifted keys to actually use the 
	capitalized versions of the letters.
	TPR 2/18/99: add the plain ascii values back in for those VMs that don't return the shifted values."

	"TextEditor initialize"
	
	| cmds |

	self initializeBasicShiftCmdKeyShortcuts.
	shiftCmdActions at: (45 + 1) put: #changeEmphasis:.		"cmd-sh-minus"
	shiftCmdActions at: (61 + 1) put: #changeEmphasis:.		"cmd-sh-plus"

	cmds := #(
		$k	changeStyle:
	).
	1 to: cmds size by: 2 do: [ :i |
		shiftCmdActions at: ((cmds at: i) asciiValue + 1) put: (cmds at: i + 1).			"plain keys"
		shiftCmdActions at: ((cmds at: i) asciiValue - 32 + 1) put: (cmds at: i + 1).		"shifted keys"
		shiftCmdActions at: ((cmds at: i) asciiValue - 96 + 1) put: (cmds at: i + 1).		"ctrl keys"
	].! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'jmv 7/21/2009 12:35'!
initializeShiftedYellowButtonMenu
	"Initialize the yellow button pop-up menu and corresponding messages."

	"TextEditor initialize"
	"
	shiftedYellowButtonMenu := SelectionMenu fromArray: {
		{'set font... (k)' translated.					#offerFontMenu}.
		{'set style... (K)' translated.				#changeStyle}.
		{'set alignment...' translated.				#chooseAlignment}.
		#-.
		{'more...' translated.						#yellowButtonActivity}.
	}
	"
	shiftedYellowButtonMenu := yellowButtonMenu! !

!TextEditor class methodsFor: 'keyboard shortcut tables' stamp: 'StephaneDucasse 7/2/2010 18:27'!
initializeYellowButtonMenu
	"Initialize the yellow button pop-up menu and corresponding messages."

	"TextEditor initialize"

	yellowButtonMenu := MenuMorph new  defaultTarget: self.
	yellowButtonMenu
		addList: {
		{'find...(f)' translated.				#find}.
		{'find again (g)' translated.			#findAgain}.
		{'set search string (h)' translated.	#setSearchString}.
		#-.
		{'do again (j)' translated.			#again}.
		{'undo (z)' translated.				#undo}.
		#-.
		{'copy (c)' translated.				#copySelection}.
		{'cut (x)' translated.					#cut}.
		{'paste (v)' translated.				#paste}.
		{'paste...' translated.					#pasteRecent}.
		#-.
		{'set font... (k)' translated.			#offerFontMenu}.
		{'set style... (K)' translated.			#changeStyle}.
		{'set alignment...' translated.			#chooseAlignment}.
		"
		#-.
		{'more...' translated.					#shiftedTextPaneMenuRequest}.
		"
	}! !

!TextEditor class methodsFor: 'accessing' stamp: 'jmv 7/21/2009 10:17'!
shiftCmdActions
	^shiftCmdActions! !

!TextEditor class methodsFor: 'class initialization' stamp: 'jmv 7/21/2009 10:37'!
shiftedYellowButtonMenu
	"Answer the menu to be presented when the yellow button is pressed while the shift key is down"

	^ shiftedYellowButtonMenu! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 11:28'!
walkAlongDisplayedLine
	^ WalkAlongDisplayedLine ifNil: [WalkAlongDisplayedLine := true]! !

!TextEditor class methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 11:29'!
walkAlongDisplayedLine: aBoolean
	WalkAlongDisplayedLine := aBoolean! !

!TextEditor class methodsFor: 'class initialization' stamp: 'jmv 7/21/2009 10:34'!
yellowButtonMenu

	^ yellowButtonMenu! !

!TextEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 3/31/2010 22:07'!
accept
	"Save the current text of the text being edited as the current acceptable version for purposes of canceling.  Allow my morph to take appropriate action"
	
	morph acceptContents! !

!TextEditor methodsFor: 'menu messages' stamp: 'FernandoOlivero 6/6/2010 16:10'!
accept: aKeyboardEvent 
	"Save the current text of the text being edited as the current acceptable version for purposes of canceling.  Allow my morph to take appropriate action"
	morph acceptContents.
	^ true! !

!TextEditor methodsFor: 'typing support' stamp: 'jmv 9/2/2009 07:23'!
addString: aString
	"Think of a better name"

	self zapSelectionWith: 
		(Text string: aString attributes: emphasisHere)! !

!TextEditor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 11:51'!
afterSelectionInsertAndSelect: aString

	self insertAndSelect: aString at: self stopIndex ! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
again
	"Text substitution. If the left shift key is down, the substitution is made 
	throughout the entire Paragraph. Otherwise, only the next possible 
	substitution is made.
	Undoer & Redoer: #undoAgain:andReselect:typedKey:."

	"If last command was also 'again', use same keys as before"
	self againOrSame: (UndoMessage sends: #undoAgain:andReselect:typedKey:)! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 4/15/2010 11:35'!
againOnce: indices
	"Find the next occurrence of FindText.  If none, answer false.
	Append the start index of the occurrence to the stream indices, and, if
	ChangeText is not the same object as FindText, replace the occurrence by it.
	Note that the search is case-sensitive for replacements, otherwise not."

	| where |
	where := paragraph text findString: FindText startingAt: self stopIndex
				caseSensitive: ((ChangeText ~~ FindText) or: [self caseSensitiveFinds]).
	where = 0 ifTrue: [^ false].
	self deselect; selectInvisiblyFrom: where to: where + FindText size - 1.
	ChangeText ~~ FindText ifTrue: [self zapSelectionWith: ChangeText].
	indices nextPut: where.
	^ true! !

!TextEditor methodsFor: 'private' stamp: 'jmv 9/2/2009 13:31'!
againOrSame: useOldKeys
	"Subroutine of search: and again.  If useOldKeys, use same FindText and ChangeText as before.
	 1/26/96 sw: real worked moved to againOrSame:many:"

	self againOrSame: useOldKeys many: false.

	(morph respondsTo: #editView) 
		ifTrue: [morph editView selectionInterval: self selectionInterval]! !

!TextEditor methodsFor: 'private' stamp: 'jmv 9/2/2009 13:26'!
againOrSame: useOldKeys many: many
	"Subroutine of search: and again.  If useOldKeys, use same FindText and ChangeText as before.  If many is true, do it repeatedly.  Created 1/26/96 sw by adding the many argument to #againOrSame."

	|  home indices wasTypedKey |

	home := self selectionInterval.  "what was selected when 'again' was invoked"

	"If new keys are to be picked..."
	useOldKeys ifFalse: "Choose as FindText..."
		[FindText := UndoSelection.  "... the last thing replaced."
		"If the last command was in another paragraph, ChangeText is set..."
		paragraph == UndoParagraph ifTrue: "... else set it now as follows."
			[UndoInterval ~= home ifTrue: [self selectInterval: UndoInterval]. "blink"
			ChangeText := ((UndoMessage sends: #undoCutCopy:) and: [self hasSelection])
				ifTrue: [FindText] "== objects signal no model-locking by 'undo copy'"
				ifFalse: [self selection]]]. "otherwise, change text is last-replaced text"

	(wasTypedKey := FindText size = 0)
		ifTrue: "just inserted at a caret"
			[home := self selectionInterval.
			self replaceSelectionWith: self nullText.  "delete search key..."
			FindText := ChangeText] "... and search for it, without replacing"
		ifFalse: "Show where the search will start"
			[home last = self selectionInterval last ifFalse:
				[self selectInterval: home]].

	"Find and Change, recording start indices in the array"
	indices := WriteStream on: (Array new: 20). "an array to store change locs"
	[(self againOnce: indices) & many] whileTrue. "<-- this does the work"
	indices isEmpty ifTrue:  "none found"
		[self flash.
		wasTypedKey ifFalse: [^self]].

	(many | wasTypedKey) ifFalse: "after undo, select this replacement"
		[home := self startIndex to:
			self startIndex + UndoSelection size - 1].

	self undoer: #undoAgain:andReselect:typedKey: with: indices contents with: home with: wasTypedKey.

	(morph respondsTo: #editView) 
		ifTrue: [morph editView selectionInterval: self selectionInterval]! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
align
	"Align text according to the next greater alignment value,
	cycling among leftFlush, rightFlush, center, and justified."
	self changeAlignment.
	self recomputeInterval! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:42'!
align: aKeyboardEvent
	"Triggered by Cmd-u;  cycle through alignment alternatives.  8/11/96 sw"

	self align.
	^ true! !

!TextEditor methodsFor: 'private' stamp: 'jmv 9/21/2009 11:25'!
applyAttribute: aTextAttribute
	"The user selected aTextAttribute via shortcut, menu or other means.
	If there is a selection, apply the attribute to the selection.
	In any case use the attribute for the user input (emphasisHere)
	It really looks like the 2 branches in this method should be closer!! (handling of Undo is different, and likely wrong)"

	| thisSel interval |

	emphasisHere := Text addAttribute: aTextAttribute toArray: emphasisHere.

	aTextAttribute isParagraphAttribute
		ifTrue: [
			interval := paragraph text encompassParagraph: self selectionInterval.
			paragraph 
				replaceFrom: interval first 
				to: interval last 
				with: ((paragraph text copyFrom: interval first to: interval last)
						addAttribute: aTextAttribute).
			paragraph composeAll.
			self recomputeSelection ]
		
		ifFalse: [
			thisSel := self selection.
			thisSel isEmpty
				ifTrue: [ paragraph composeAll ]
				ifFalse: [
					self replaceSelectionWith: (thisSel addAttribute: aTextAttribute).
					paragraph composeAll.
					self recomputeSelection] ].

	morph possiblyChanged ! !

!TextEditor methodsFor: 'typing support' stamp: 'FernandoOlivero 3/13/2010 17:09'!
backTo: startIndex
	"During typing, backspace to startIndex.  Deleted characters fall into three
	 clusters, from left to right in the text: (1) preexisting characters that were
	 backed over; (2) newly typed characters that were backed over;
	(3) preexisting characters that
	 were highlighted before typing began.  If typing has not yet been opened,
	 open it and watch for the first and third cluster.  If typing has been opened,
	 watch for the first and second cluster.  Save characters from the first and third
	 cluster in UndoSelection.  Tally characters from the first cluster in UndoMessage's parameter.
	 Delete all the clusters.  Do not alter Undoer or UndoInterval (except via
	 openTypeIn).  The code is shorter than the comment."

	| saveLimit newBackovers |
	saveLimit := beginTypeInBlock
		ifNil: [self openTypeIn. UndoSelection := self nullText. self stopIndex]
		ifNotNil: [self startOfTyping].
	self markIndex: startIndex.
	startIndex < saveLimit ifTrue: [
		newBackovers := self startOfTyping - startIndex.
		beginTypeInBlock := self startIndex.
		UndoSelection replaceFrom: 1 to: 0 with:
			(paragraph text copyFrom: startIndex to: saveLimit - 1).
		UndoMessage numArgs > 0 ifTrue: [ 
			UndoMessage argument: (UndoMessage argument ifNil: [1]) + newBackovers]].
	self zapSelectionWith: self nullText.
	self unselect! !

!TextEditor methodsFor: 'binding' stamp: 'jmv 11/4/2008 11:51'!
bindingOf: aString
	^model bindingOf: aString! !

!TextEditor methodsFor: 'editing keys' stamp: 'FernandoOlivero 5/17/2010 15:00'!
blankParagraph
	| blankParagraph |
	blankParagraph  := MultiNewParagraph new.
	blankParagraph 
		compose:  '' asText 
		style:  TextStyle default copy    
		from: 1
		in:  (0@0 extent: 10@10) .
	^ blankParagraph	! !

!TextEditor methodsFor: 'parenblinking' stamp: 'jmv 8/18/2009 22:32'!
blinkParen
	lastParenLocation ifNotNil: [
		self text string size >= lastParenLocation ifTrue: [
			self text
				addAttribute: TextEmphasis bold
				from: lastParenLocation
				to: lastParenLocation]]! !

!TextEditor methodsFor: 'parenblinking' stamp: 'jmv 11/4/2008 11:51'!
blinkParenAt: parenLocation 
	self text
		addAttribute: TextEmphasis bold
		from: parenLocation
		to: parenLocation.
	lastParenLocation := parenLocation.! !

!TextEditor methodsFor: 'parenblinking' stamp: 'jmv 9/2/2009 13:20'!
blinkPrevParen: aKeyboardEvent
	| openDelimiter closeDelimiter level string here hereChar |
	string := paragraph text string.
	here := pointBlock stringIndex.
	openDelimiter := aKeyboardEvent keyValue.
	closeDelimiter := '([{' at: (')]}' indexOf: openDelimiter).
	level := 1.
	[level > 0 and: [here > 2]]
		whileTrue:
			[hereChar := string at: (here := here - 1).
			hereChar = closeDelimiter
				ifTrue:
					[level := level - 1.
					level = 0
						ifTrue: [^ self blinkParenAt: here]]
				ifFalse:
					[hereChar = openDelimiter
						ifTrue: [level := level + 1]]]! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
cancel
	"Cancel the changes made so far to this text"
	morph cancelEdits! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:42'!
cancel: aKeyboardEvent
	"Cancel unsubmitted changes."

	self cancel.
	^ true! !

!TextEditor methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 11:34'!
caseSensitiveFinds
	^ self class caseSensitiveFinds! !

!TextEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 7/2/2010 18:46'!
changeAlignment
	| aList reply  |
	aList := #(leftFlush centered justified rightFlush).
	reply := UIManager default chooseFrom:  aList values: aList.	
	reply ifNil:[^self].
	self applyAttribute: (TextAlignment perform: reply).
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'StephaneDucasse 7/2/2010 18:54'!
changeEmphasis: aKeyboardEvent
	"Change the emphasis of the current selection or prepare to accept characters with the change in emphasis. Emphasis change amounts to a font change.  Keeps typeahead."

	"control 0..9 -> 0..9"

	| keyCode attribute oldAttributes index thisSel colors |
	keyCode := ('0123456789-=' indexOf: aKeyboardEvent keyCharacter ifAbsent: [1]) - 1.
	oldAttributes := paragraph text attributesAt: self pointIndex.
	thisSel := self selection.

	"Decipher keyCodes for Command 0-9..."
	"
	(keyCode between: 1 and: 5) 
		ifTrue: [attribute := TextFontChange fontNumber: keyCode].
	"
	keyCode = 6 
		ifTrue: [
			colors := #(#black #magenta #red #yellow #green #blue #cyan #white).
			index := UIManager default chooseFrom:  colors title: 'choose color...'.
			index = 0 ifTrue: [^true].
			index <= colors size 
				ifTrue: [attribute := TextColor color: (Color perform: (colors at: index))]
				ifFalse: [
					index := index - colors size - 1.	"Re-number!!!!!!"
					index = 0 ifTrue: [attribute := self chooseColor].
					thisSel ifNil: [^true]	"Could not figure out what to link to"]].
	(keyCode between: 7 and: 11) 
		ifTrue: [
			aKeyboardEvent shiftPressed 
				ifTrue: [
					keyCode = 10 ifTrue: [attribute := TextKern kern: -1].
					keyCode = 11 ifTrue: [attribute := TextKern kern: 1]]
				ifFalse: [
					attribute := TextEmphasis 
								perform: (#(#bold #italic #narrow #underlined #struckOut) at: keyCode - 6).
					oldAttributes 
						do: [:att | (att dominates: attribute) ifTrue: [attribute turnOff]]]].
	keyCode = 0 ifTrue: [ attribute := TextEmphasis normal ].
	attribute ifNotNil: [
		self applyAttribute: attribute].
	^true! !

!TextEditor methodsFor: 'attributes' stamp: 'StephaneDucasse 7/2/2010 18:47'!
changeEmphasisOrAlignment
	| aList reply  code align menuList startIndex attribute |
	startIndex := self startIndex.
	aList := #(normal bold italic narrow underlined struckOut leftFlush centered rightFlush justified).	
	align := paragraph text alignmentAt: startIndex.
	code := paragraph text emphasisAt: startIndex.
	menuList := WriteStream on: Array new.
	menuList nextPut: (code isZero ifTrue:['<on>'] ifFalse:['<off>']), 'normal' translated.
	menuList nextPutAll: (#(bold italic underlined struckOut) collect:[:emph|
		(code anyMask: (TextEmphasis perform: emph) emphasisCode)
			ifTrue: [ '<on>', emph asString translated ]
			ifFalse: [ '<off>',emph asString translated ]]).
	((paragraph text attributesAt: startIndex)
		anySatisfy: [ :attr | attr isKern and: [attr kern < 0 ]]) 
			ifTrue: [ menuList nextPut:'<on>', 'narrow' translated ]
			ifFalse: [ menuList nextPut:'<off>', 'narrow' translated ].
	menuList nextPutAll: (#(leftFlush centered rightFlush justified) collectWithIndex: [ :type :i |
		align = (i-1)
			ifTrue: [ '<on>',type asString translated ]
			ifFalse: [ '<off>',type asString translated ]]).
	aList := #(normal bold italic underlined struckOut narrow leftFlush centered rightFlush justified).
	reply := UIManager default chooseFrom:  aList values: aList.
	reply notNil ifTrue: [
		(#(leftFlush centered rightFlush justified) includes: reply)
			ifTrue: [
				attribute :=TextAlignment perform: reply]
			ifFalse: [
				attribute := TextEmphasis perform: reply].
		self applyAttribute:  attribute].
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:42'!
changeLfToCr: aKeyboardEvent
	"Replace all LFs by CRs.
	Triggered by Cmd-U -- useful when getting code from FTP sites
	jmv- Modified to als change crlf by cr"
	
	| fixed |
	fixed := self selection string.
	fixed := fixed copyReplaceAll: String crlf with: String cr.
	fixed := fixed copyReplaceAll: String lf with: String cr. 
	self replaceSelectionWith: (Text fromString: fixed).
	^ true! !

!TextEditor methodsFor: 'initialize-release' stamp: 'FernandoOlivero 3/13/2010 15:32'!
changeParagraph: aParagraph 
	"Install aParagraph as the one to be edited by the receiver."

	UndoParagraph == paragraph ifTrue: [UndoParagraph := nil].
	paragraph := aParagraph.
	self resetState! !

!TextEditor methodsFor: 'attributes' stamp: 'StephaneDucasse 7/2/2010 18:33'!
changeStyle
	"Let user change styles for the current text pane."
	| names reply style current menuList |

	current := paragraph text initialStyle.
	names := TextStyle availableTextStyleNames.
	menuList := names collect: [ :styleName |
		styleName = current name
			ifTrue: [ '<on>', styleName ]
			ifFalse: [ '<off>', styleName ]].
	reply := UIManager default chooseFrom:  menuList values: names.	
	reply ifNotNil: [
		(style := TextStyle named: reply) ifNil: [Beeper beep. ^ true].
		paragraph text initialStyle: style.
		paragraph composeAll.
		self recomputeSelection].
	^ true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:42'!
changeStyle: aKeyboardEvent 
	"Put up the style-change menu"

	self closeTypeIn.
	self changeStyle.
	^ true! !

!TextEditor methodsFor: 'attributes' stamp: 'jmv 9/1/2009 22:56'!
changeTextFont
	"Present a menu of available fonts, and if one is chosen, apply it to the current selection."
	| curFont newFont attr startIndex |
	startIndex := self startIndex.
	curFont := paragraph text fontAt: startIndex.
	newFont := StrikeFont fromUser: curFont.
	newFont ifNil: [ ^self ].
	attr := TextFontReference toFont: newFont.
	self applyAttribute: attr! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
chooseAlignment
	self changeAlignment! !

!TextEditor methodsFor: 'editing keys' stamp: 'AlainPlantec 4/14/2010 15:46'!
chooseColor
	"Make a new Text Color Attribute, let the user pick a color, and return the attribute"

	| attribute |	
	attribute := TextColor color: Color black "default".
	UIManager default chooseColor
		ifNotNil: [:nc | attribute color: nc].
	^ attribute
! !

!TextEditor methodsFor: 'parenblinking' stamp: 'jmv 8/16/2009 18:31'!
clearParens
	lastParenLocation ifNotNil: [
		self text string size >= lastParenLocation ifTrue: [
			self text
				removeAttribute: TextEmphasis bold
				from: lastParenLocation
				to: lastParenLocation]].
	lastParenLocation := nil! !

!TextEditor methodsFor: 'typing support' stamp: 'jmv 9/2/2009 11:48'!
closeTypeIn
	"See comment in openTypeIn.  It is important to call closeTypeIn before executing
	 any non-typing key, making a new selection, etc.  It is called automatically for
	 menu commands.
	 Undoer & Redoer: undoAndReselect:redoAndReselect:."

	| begin stop |
	beginTypeInBlock == nil ifFalse: [
		(UndoMessage sends: #noUndoer) ifTrue: "should always be true, but just in case..."
			[begin := self startOfTyping.
			stop := self stopIndex.
			self undoer: #undoAndReselect:redoAndReselect:
				with: (begin + UndoMessage argument to: begin + UndoSelection size - 1)
				with: (stop to: stop - 1).
			UndoInterval := begin to: stop - 1].
		beginTypeInBlock := nil]! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
compareToClipboard
	"Check to see if whether the receiver's text is the same as the text currently on the clipboard, and inform the user."
	| s1 s2 |
	s1 := self clipboardText string.
	s2 := paragraph text string.
	s1 = s2 ifTrue: [^ self inform: 'Exact match'].

	(StringHolder new textContents:
		(TextDiffBuilder buildDisplayPatchFrom: s1 to: s2))
		openLabel: 'Comparison to Clipboard Text'! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:42'!
compareToClipboard: aKeyboardEvent
	"Compare the receiver to the text on the clipboard."

	self compareToClipboard.
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 11/4/2008 11:51'!
copyHiddenInfo
	"In TextLinks, TextDoits, TextColor, and TextURLs, there is hidden
info.  Copy that to the clipboard.  You can paste it and see what it is.
Usually enclosed in <>."

	^ self clipboardTextPut: self hiddenInfo asText! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
copySelection
	"Copy the current selection and store it in the paste buffer, unless a caret.  Undoer & Redoer: undoCutCopy"

	self lineSelectAndEmptyCheck: [^ self].

	"Simulate 'substitute: self selection' without locking the controller"
	UndoSelection := self selection.
	self undoer: #undoCutCopy: with: self clipboardText.
	UndoInterval := self selectionInterval.
	self clipboardTextPut: UndoSelection! !

!TextEditor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 11:51'!
correctFrom: start to: stop with: aString
	"Make a correction in the model that the user has authorised from somewhere else in the system (such as from the compilier).  The user's selection is not changed, only corrected."
	| wasShowing userSelection delta loc |
	aString = '#insert period' ifTrue:
		[loc := start.
		[(loc := loc-1)>0 and: [(paragraph text string at: loc) isSeparator]]
			whileTrue: [loc := loc-1].
		^ self correctFrom: loc+1 to: loc with: '.'].
	(wasShowing := selectionShowing) ifTrue: [ self reverseSelection ].
	userSelection := self selectionInterval.

	self selectInvisiblyFrom: start to: stop.
	self replaceSelectionWith: aString asText.

	delta := aString size - (stop - start + 1).
	self selectInvisiblyFrom:
		userSelection first + (userSelection first > start ifFalse: [ 0 ] ifTrue: [ delta ])
		to: userSelection last + (userSelection last > start ifFalse: [ 0 ] ifTrue: [ delta ]).
	wasShowing ifTrue: [ self reverseSelection ].
! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:43'!
crWithIndent: aKeyboardEvent
	"Replace the current text selection with CR followed by as many tabs
	as on the current line (+/- bracket count) -- initiated by Shift-Return."
	| char s i tabCount |
	s := paragraph string.
	i := self stopIndex.
	tabCount := 0.
	[(i := i-1) > 0 and: [(char := s at: i) ~= Character cr]]
		whileTrue:  "Count tabs and brackets (but not a leading bracket)"
		[(char = Character tab and: [i < s size and: [(s at: i+1) ~= $[ ]]) ifTrue: [tabCount := tabCount + 1].
		char = $[ ifTrue: [tabCount := tabCount + 1].
		char = $] ifTrue: [tabCount := tabCount - 1]].
	 "Now inject CR with tabCount tabs"
	self addString: (String streamContents: [ :strm | strm crtab: tabCount ]).
	^ false! !

!TextEditor methodsFor: 'accessing' stamp: 'StephaneDucasse 3/31/2010 22:08'!
currentAttributes
	| i t |
	i := self startIndex.
	t := paragraph text.
	^i > t size
		ifTrue: [ emphasisHere ]
		ifFalse: [ "paragraph attributes at the current paragraph will later be applied.
			include them in the answer"
			(emphasisHere reject: [ :attr | attr isParagraphAttribute ]),
				((t attributesAt: i) select: [ :attr | attr isParagraphAttribute ]) ]! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 9/21/2009 09:55'!
currentCharacterStyleOrNil
	"Answer the CharacterStyle for the current selection or cursor location if any"

	^self hasSelection
		ifTrue: [ paragraph text characterStyleOrNilAt: self startIndex ]
		ifFalse: [ paragraph text characterStyleOrNilIfApplying: emphasisHere ]! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 9/21/2009 10:01'!
currentParagraphStyle
	"Answer the TextStyle for the current selection or cursor location if any"
	
	| i t |
	i := self startIndex.
	t := paragraph text.
	^i > t size
		ifTrue: [ paragraph text styleIfApplying: emphasisHere ]
		ifFalse: [ paragraph text textStyleAt: i ]! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'FernandoOlivero 5/17/2010 15:06'!
cursorEnd: aKeyboardEvent 

	"Private - Move cursor end of current line."
	| string |
	self closeTypeIn.
	string := paragraph text string.
	self
		moveCursor:
			[:position | self walkAlongDisplayedLine
				ifTrue:[| targetLine |
					targetLine := paragraph lines at:(paragraph lineIndexOfCharacterIndex: position).
					targetLine = paragraph lines  last 
						ifTrue:[targetLine last + 1]
						ifFalse:[targetLine last]]
				ifFalse:[
					string
						indexOf: Character cr
						startingAt: position
						ifAbsent:[string size + 1]]]
		forward: true
		specialBlock:[:dummy | string size + 1]
		event: aKeyboardEvent.
	^true! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'AlainPlantec 4/15/2010 11:33'!
cursorHome: aKeyboardEvent

	"Private - Move cursor from position in current line to beginning of
	current line. If control key is pressed put cursor at beginning of text"

	| string |
	string := paragraph text string.
	self
		moveCursor: [ :position | self walkAlongDisplayedLine
				ifTrue:[
					(paragraph lines at:(paragraph lineIndexOfCharacterIndex: position)) first]
				ifFalse:[
					(string
						lastIndexOf: Character cr
						startingAt: position - 1
						ifAbsent:[0]) + 1]]
		forward: false
		specialBlock: [:dummy | 1]
		event: aKeyboardEvent.
	^true! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
cut
	"Cut out the current selection and redisplay the paragraph if necessary.  Undoer & Redoer: undoCutCopy:"

	self lineSelectAndEmptyCheck: [^ self].

	self replaceSelectionWith: self nullText. 
	self undoer: #undoCutCopy: with: self clipboardText.
	self clipboardTextPut: UndoSelection! !

!TextEditor methodsFor: 'typing support' stamp: 'FernandoOlivero 6/6/2010 15:54'!
dispatchOn: aKeyboardEvent
	"Carry out the action associated with this character, if any.
	Type-ahead is passed so some routines can flush or use it."

	| asciiValue honorCommandKeys |
	asciiValue := aKeyboardEvent keyValue.
	asciiValue = Character cr asciiValue
		ifTrue:[
	 		morph acceptOnCR
				ifTrue: [
					self closeTypeIn.
					^ true].
			self clearParens.
			aKeyboardEvent controlKeyPressed ifTrue: [ ^ self normalCharacter: aKeyboardEvent ].
			aKeyboardEvent shiftPressed ifTrue: [ ^ self lf: aKeyboardEvent ].
			aKeyboardEvent commandKeyPressed ifTrue: [ ^ self crlf: aKeyboardEvent ].
			^ self crWithIndent: aKeyboardEvent ] .

	honorCommandKeys := self cmdKeysInText.

	"Special keys overwrite crtl+key combinations - at least on Windows. To resolve this
	conflict, assume that keys other than cursor keys aren't used together with Crtl." 
	((self class specialShiftCmdKeys includes: asciiValue) and: [ asciiValue < 27])
		ifTrue: [^ aKeyboardEvent controlKeyPressed
			ifTrue: [self perform: (self class shiftCmdActions at: asciiValue + 1) with: aKeyboardEvent]
			ifFalse: [self perform: (self class cmdActions at: asciiValue + 1) with: aKeyboardEvent]].

	"backspace, and escape keys (ascii 8 and 27) are command keys"
	((honorCommandKeys and: [aKeyboardEvent commandKeyPressed]) or: [self class specialShiftCmdKeys includes: asciiValue]) ifTrue: [
		^  aKeyboardEvent shiftPressed
			ifTrue: [
				self perform: (self class shiftCmdActions at: asciiValue + 1) with: aKeyboardEvent]
			ifFalse: [
				self perform: (self class cmdActions at: asciiValue + 1) with: aKeyboardEvent]].

	"the control key can be used to invoke shift-cmd shortcuts"
	(honorCommandKeys and: [aKeyboardEvent controlKeyPressed])
		ifTrue: [
			^ self perform: (self class shiftCmdActions at: asciiValue + 1) with: aKeyboardEvent].

	(')]}' includes: aKeyboardEvent keyValue)
		ifTrue: [ self blinkPrevParen: aKeyboardEvent ].

	^ self normalCharacter: aKeyboardEvent! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:43'!
doAgainMany: aKeyboardEvent 
	"Do the previous thing again repeatedly. 1/26/96 sw"

	self closeTypeIn.
	self againOrSame: (UndoMessage sends: #undoAgain:andReselect:typedKey:) many: true.
	^ true! !

!TextEditor methodsFor: 'typing support' stamp: 'jmv 11/4/2008 11:51'!
doneTyping
	beginTypeInBlock := nil! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:43'!
duplicate: aKeyboardEvent
	"Paste the current selection over the prior selection, if it is non-overlapping and
	 legal.  Undoer & Redoer: undoAndReselect."

	self closeTypeIn.
	(self hasSelection and: [self isDisjointFrom: otherInterval])
		ifTrue: "Something to duplicate"
			[self replace: otherInterval with: self selection and:
				[self selectAt: self pointIndex]]
		ifFalse:
			[morph flash].
	^true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:43'!
enclose: aKeyboardEvent
	"Insert or remove bracket characters around the current selection."

	| left right startIndex stopIndex oldSelection which text |
	self closeTypeIn.
	startIndex := self startIndex.
	stopIndex := self stopIndex.
	oldSelection := self selection.
	which := '([<{"''' indexOf: aKeyboardEvent keyCharacter ifAbsent: [ ^true ].
	left := '([<{"''' at: which.
	right := ')]>}"''' at: which.
	text := paragraph text.
	((startIndex > 1 and: [stopIndex <= text size])
			and: [ (text at: startIndex-1) = left and: [(text at: stopIndex) = right]])
		ifTrue: [
			"already enclosed; strip off brackets"
			self selectFrom: startIndex-1 to: stopIndex.
			self replaceSelectionWith: oldSelection]
		ifFalse: [
			"not enclosed; enclose by matching brackets"
			self replaceSelectionWith:
				(Text string: (String with: left), oldSelection string, (String with: right) attributes: emphasisHere).
			self selectFrom: startIndex+1 to: stopIndex].
	^true! !

!TextEditor methodsFor: 'private' stamp: 'FernandoOlivero 5/17/2010 13:52'!
encompassParagraph: anInterval
	"Return an interval that includes anInterval, and that comprises one or several whole paragraphs in the receiver.
	Answer starts at the position following a cr (or eventually 1) and ends at a cr (or eventually at self size)"
	| left right string |
	string := paragraph string .
	left := (string lastIndexOf: Character cr startingAt: anInterval first - 1 ifAbsent:[0]) + 1.
	right := (string indexOf: Character cr startingAt: (anInterval last max: anInterval first) ifAbsent: [string size]).
	^left to: right! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
exchange
	"See comment in exchangeWith:"

	self exchangeWith: otherInterval! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:43'!
exchange: aKeyboardEvent
	"Exchange the current and prior selections."

	self closeTypeIn.
	self exchange.
	^true! !

!TextEditor methodsFor: 'private' stamp: 'jmv 11/4/2008 11:51'!
exchangeWith: prior
	"If the prior selection is non-overlapping and legal, exchange the text of
	 it with the current selection and leave the currently selected text selected
	 in the location of the prior selection (or leave a caret after a non-caret if it was
	 exchanged with a caret).  If both selections are carets, flash & do nothing.
	 Don't affect the paste buffer.  Undoer: itself; Redoer: Undoer."

	| start stop before selection priorSelection delta altInterval |
	start := self startIndex.
	stop := self stopIndex - 1.
	((prior first <= prior last) | (start <= stop) "Something to exchange" and:
			[self isDisjointFrom: prior])
		ifTrue:
			[before := prior last < start.
			selection := self selection.
			priorSelection := paragraph text copyFrom: prior first to: prior last.

			delta := before ifTrue: [0] ifFalse: [priorSelection size - selection size].
			self zapSelectionWith: priorSelection.
			self selectFrom: prior first + delta to: prior last + delta.

			delta := before ifTrue: [stop - prior last] ifFalse: [start - prior first].
			self zapSelectionWith: selection.
			altInterval := prior first + delta to: prior last + delta.
			self undoer: #exchangeWith: with: altInterval.
			"If one was a caret, make it otherInterval & leave the caret after the other"
			prior first > prior last ifTrue: [self selectAt: UndoInterval last + 1].
			otherInterval := start > stop
				ifTrue: [self selectAt: altInterval last + 1. UndoInterval]
				ifFalse: [altInterval]]
		ifFalse:
			[morph flash]! !

!TextEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 7/2/2010 18:41'!
find
	"Prompt the user for a string to search for, and search the receiver from the current selection onward for it."

	| reply |
	reply := UIManager default request: 'Find what? ' initialAnswer: ''.
	reply size = 0 ifTrue: [
		^ self].
	self setSearch: reply.
	ChangeText := FindText.  "Implies no replacement to againOnce: method"
	self againOrSame: true.

	morph installEditorToReplace: self! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 12:43'!
find: aKeyboardEvent
	"Prompt the user for what to find, then find it, searching from the current selection onward.  1/24/96 sw"

	self closeTypeIn.
	self find.
	^ true! !

!TextEditor methodsFor: 'menu messages' stamp: 'StephaneDucasse 7/2/2010 18:41'!
findAgain
	"Find the text-to-find again."

	self againOrSame: true! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 13:30'!
findAgain: aKeyboardEvent 
	"Find the desired text again.  1/24/96 sw"

	self closeTypeIn.
	self againOrSame: true many: aKeyboardEvent shiftPressed.
	^ true! !

!TextEditor methodsFor: 'displaying' stamp: 'jmv 11/4/2008 11:51'!
flash
	^ morph flash! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 13:03'!
forwardDelete: aKeyboardEvent
	"Delete forward over the next character.
	  Make Undo work on the whole type-in, not just the one char.
	wod 11/3/1998: If there was a selection use #zapSelectionWith: rather than #backspace: which was 'one off' in deleting the selection. Handling of things like undo or typeIn area were not fully considered."
	| startIndex usel upara uinterval ind stopIndex |
	startIndex := self markIndex.
	startIndex > paragraph text size ifTrue: [
		^ false].
	self hasSelection ifTrue: [
		"there was a selection"
		self zapSelectionWith: self nullText.
		^ false].
	"Null selection - do the delete forward"
	beginTypeInBlock == nil	"no previous typing.  openTypeIn"
		ifTrue: [self openTypeIn. UndoSelection := self nullText].
	uinterval := UndoInterval deepCopy.
	upara := UndoParagraph deepCopy.
	stopIndex := startIndex.
	(aKeyboardEvent keyValue = 127 and: [ aKeyboardEvent shiftPressed ])
		ifTrue: [stopIndex := (self nextWord: stopIndex) - 1].
	self selectFrom: startIndex to: stopIndex.
	self replaceSelectionWith: self nullText.
	self selectFrom: startIndex to: startIndex-1.
	UndoParagraph := upara.  UndoInterval := uinterval.
	UndoMessage selector == #noUndoer ifTrue: [
		(UndoSelection isText) ifTrue: [
			usel := UndoSelection.
			ind := startIndex. "UndoInterval startIndex"
			usel replaceFrom: usel size + 1 to: usel size with:
				(UndoParagraph text copyFrom: ind to: ind).
			UndoParagraph text replaceFrom: ind to: ind with:
self nullText]].
	^false! !

!TextEditor methodsFor: 'private' stamp: 'jmv 11/4/2008 11:51'!
getPluggableYellowButtonMenu: shiftKeyState
	| customMenu |
	^ (morph notNil and: [(customMenu := morph getMenu: shiftKeyState) notNil])
		ifTrue: [customMenu]
		ifFalse:
			[shiftKeyState
				ifTrue: [self class shiftedYellowButtonMenu]
				ifFalse: [self class yellowButtonMenu]]! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
hasCaret
	^self markBlock = self pointBlock! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/1/2009 15:47'!
hiddenInfo
	"In TextLinks, TextDoits, TextColor, and TextURLs, there is hidden info.  Return the entire string that was used by Cmd-6 to create this text attribute.  Usually enclosed in < >."

	| attrList |
	attrList := paragraph text attributesAt: (self pointIndex + self markIndex)//2.
	attrList do: [:attr |
		attr forTextActionInfoDo: [ :info |
			^ self selection asString, '<', info, '>']].
	"If none of the above"
	attrList do: [:attr |
		attr forTextColorDo: [ :color |
			^ self selection asString, '<', color printString, '>']].
	^ self selection asString, '[No hidden info]'! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 12/29/2009 14:42'!
inOutdent: aKeyboardEvent delta: delta
	"Add/remove a tab at the front of every line occupied by the selection. 
	Derived from work by Larry Tesler back in December 1985.  Now triggered by Cmd-L and Cmd-R.  2/29/96 sw"

	| cr realStart realStop lines startLine stopLine start stop adjustStart indentation size numLines inStream newString outStream |
	cr := Character cr.

	"Operate on entire lines, but remember the real selection for re-highlighting later"
	realStart := self startIndex.
	realStop := self stopIndex - 1.

	"Special case a caret on a line of its own, including weird case at end of paragraph"
	(realStart > realStop and: [
				realStart < 2 or: [(paragraph string at: realStart - 1) == cr]])
		ifTrue:
			[delta < 0
				ifTrue: [
					morph flash]
				ifFalse: [
					self replaceSelectionWith: Character tab asSymbol asText.
					self selectAt: realStart + 1].
			^true].

	lines := paragraph lines.
	startLine := paragraph lineIndexOfCharacterIndex: realStart.
	stopLine := paragraph lineIndexOfCharacterIndex: (realStart max: realStop).
	start := (lines at: startLine) first.
	stop := (lines at: stopLine) last.
	
	"Pin the start of highlighting unless the selection starts a line"
	adjustStart := realStart > start.

	"Find the indentation of the least-indented non-blank line; never outdent more"
	indentation := (startLine to: stopLine) inject: 1000 into: [ :previousValue :each |
		previousValue min: (paragraph indentationOfLineIndex: each ifBlank: [ :tabs | 1000 ])].			

	size :=  stop + 1 - start.
	numLines := stopLine + 1 - startLine.
	inStream := ReadStream on: paragraph string from: start to: stop.

	newString := String new: size + ((numLines * delta) max: 0).
	outStream := ReadWriteStream on: newString.

	"This subroutine does the actual work"
	self indent: delta fromStream: inStream toStream: outStream.

	"Adjust the range that will be highlighted later"
	adjustStart ifTrue: [realStart := (realStart + delta) max: start].
	realStop := realStop + outStream position - size.

	"Prepare for another iteration"
	indentation := indentation + delta.
	size := outStream position.
	inStream := outStream setFrom: 1 to: size.

	outStream == nil
		ifTrue: 	"tried to outdent but some line(s) were already left flush"
			[morph flash]
		ifFalse:
			[self selectInvisiblyFrom: start to: stop.
			size = newString size ifFalse: [newString := outStream contents].
			self replaceSelectionWith: newString asText].
	self selectFrom: realStart to: realStop. 	"highlight only the original range"
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:44'!
indent: aKeyboardEvent
	"Add a tab at the front of every line occupied by the selection. Invoked from keyboard via cmd-shift-R.  2/29/96 sw"

	^ self inOutdent: aKeyboardEvent delta: 1! !

!TextEditor methodsFor: 'private' stamp: 'jmv 11/4/2008 11:51'!
indent: delta fromStream: inStream toStream: outStream
	"Append the contents of inStream to outStream, adding or deleting delta or -delta
	 tabs at the beginning, and after every CR except a final CR.  Do not add tabs
	 to totally empty lines, and be sure nothing but tabs are removed from lines."

	| ch skip cr tab prev atEnd |
	cr := Character cr.
	tab := Character tab.
	delta > 0
		ifTrue: "shift right"
			[prev := cr.
			 [ch := (atEnd := inStream atEnd) ifTrue: [cr] ifFalse: [inStream next].
			  (prev == cr and: [ch ~~ cr]) ifTrue:
				[delta timesRepeat: [outStream nextPut: tab]].
			  atEnd]
				whileFalse:
					[outStream nextPut: ch.
					prev := ch]]
		ifFalse: "shift left"
			[skip := delta. "a negative number"
			 [inStream atEnd] whileFalse:
				[((ch := inStream next) == tab and: [skip < 0]) ifFalse:
					[outStream nextPut: ch].
				skip := ch == cr ifTrue: [delta] ifFalse: [skip + 1]]]! !

!TextEditor methodsFor: 'editing keys' stamp: 'FernandoOlivero 5/17/2010 14:57'!
initialize
	super initialize.
	self changeParagraph:  self blankParagraph.! !

!TextEditor methodsFor: 'new selection' stamp: 'jmv 11/4/2008 11:51'!
insertAndSelect: aString at: anInteger

	self replace: (anInteger to: anInteger - 1)
		with: (Text string: (' ' , aString)
					attributes: emphasisHere)
		and: [self ]! !

!TextEditor methodsFor: 'private' stamp: 'jmv 11/4/2008 11:51'!
isDisjointFrom: anInterval
	"Answer true if anInterval is a caret not touching or within the current
	 interval, or if anInterval is a non-caret that does not overlap the current
	 selection."

	| fudge |
	fudge := anInterval size = 0 ifTrue: [1] ifFalse: [0].
	^(anInterval last + fudge < self startIndex or:
			[anInterval first - fudge >= self stopIndex])
! !

!TextEditor methodsFor: 'undo support' stamp: 'jmv 11/4/2008 11:51'!
isDoing
	"Call from a doer/undoer/redoer any time to see which it is."

	^(self isUndoing | self isRedoing) not! !

!TextEditor methodsFor: 'undo support' stamp: 'jmv 11/4/2008 11:51'!
isRedoing
	"Call from a doer/undoer/redoer any time to see which it is."

	^UndoParagraph == #redoing! !

!TextEditor methodsFor: 'testing' stamp: 'FernandoOlivero 6/15/2010 13:07'!
isTextEditor
	^ true ! !

!TextEditor methodsFor: 'undo support' stamp: 'jmv 11/4/2008 11:51'!
isUndoing
	"Call from a doer/undoer/redoer any time to see which it is."

	^UndoParagraph == #undoing! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 9/21/2009 10:42'!
lastFont
	"Answer the Font for to be used if positioned at the end of the text"
	
	| t |
	t := paragraph text.
	^self startIndex > t size
		ifTrue: [ paragraph text fontIfApplying: emphasisHere ]
		ifFalse: [ paragraph text fontAt: t size +1 ]! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 9/21/2009 10:42'!
lastParagraphStyle
	"Answer the TextStyle for to be used if positioned at the end of the text"
	
	| t |
	t := paragraph text.
	^self startIndex > t size
		ifTrue: [ paragraph text styleIfApplying: emphasisHere ]
		ifFalse: [ paragraph text textStyleAt: t size +1 ]! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:44'!
makeCapitalized: aKeyboardEvent
	"Force the current selection to uppercase.  Triggered by Cmd-X."

	| prev |
	prev := $-.  "not a letter"
	self replaceSelectionWith: (Text fromString:
		(self selection string collect:
			[:c | prev := prev isLetter ifTrue: [c asLowercase] ifFalse: [c asUppercase]])).
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:44'!
makeLowercase: aKeyboardEvent
	"Force the current selection to lowercase.  Triggered by Cmd-X."

	self replaceSelectionWith: (Text fromString: (self selection string asLowercase)).
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:44'!
makeUppercase: aKeyboardEvent
	"Force the current selection to uppercase.  Triggered by Cmd-Y."

	self replaceSelectionWith: (Text fromString: (self selection string asUppercase)).
	^ true! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:51'!
markBlock
	^ markBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:51'!
markBlock: aCharacterBlock
	markBlock := aCharacterBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
markIndex
	^ self markBlock stringIndex! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:59'!
markIndex: anIndex
	self markBlock: (paragraph characterBlockForIndex: anIndex)
! !

!TextEditor methodsFor: 'model access' stamp: 'jmv 11/4/2008 11:51'!
model: aModel 
	"Controller|model: and Controller|view: are sent by View|controller: in 
	order to coordinate the links between the model, view, and controller. In 
	ordinary usage, the receiver is created and passed as the parameter to 
	View|controller: so that the receiver's model and view links can be set 
	up by the view."

	model := aModel! !

!TextEditor methodsFor: 'events' stamp: 'FernandoOlivero 3/13/2010 15:38'!
mouseDown: evt 
	"An attempt to break up the old processRedButton code into threee phases"
	| clickPoint b |

	oldInterval := self selectionInterval.
	clickPoint := evt cursorPoint.
	b := paragraph characterBlockAtPoint: clickPoint.

	(paragraph clickAt: clickPoint for: model controller: self ) ifTrue: [
		self markBlock: b.
		self pointBlock: b.
		evt hand releaseKeyboardFocus: self.
		^ self ].
	
	evt shiftPressed
		ifFalse: [
			self closeTypeIn.
			self markBlock: b.
			self pointBlock: b ]! !

!TextEditor methodsFor: 'events' stamp: 'jmv 8/3/2009 12:13'!
mouseMove: evt 
	"Change the selection in response to mouse-down drag"

	self pointBlock: (paragraph characterBlockAtPoint: (evt cursorPoint)).
	self storeSelectionInParagraph! !

!TextEditor methodsFor: 'events' stamp: 'jmv 9/1/2009 10:23'!
mouseUp: evt
	"An attempt to break up the old processRedButton code into threee phases"
	oldInterval ifNil: [^ self].  "Patched during clickAt: repair"
	(self hasCaret 
		and: [oldInterval = self selectionInterval])
		ifTrue: [self selectWord].
	self setEmphasisHereFromText.
	(self isDisjointFrom: oldInterval) ifTrue:
		[otherInterval := oldInterval].
	self storeSelectionInParagraph! !

!TextEditor methodsFor: 'private' stamp: 'jmv 9/2/2009 13:17'!
moveCursor: directionBlock forward: forward specialBlock: specialBlock event: aKeyboardEvent

	super moveCursor: directionBlock forward: forward specialBlock: specialBlock event: aKeyboardEvent.
	self setEmphasisHereFromTextForward: forward! !

!TextEditor methodsFor: 'undo support' stamp: 'jmv 11/4/2008 11:51'!
noUndoer
	"The Undoer to use when the command can not be undone.  Checked for
	 specially by readKeyboard."

	UndoMessage := Message selector: #noUndoer! !

!TextEditor methodsFor: 'private' stamp: 'jmv 7/29/2009 22:36'!
nullText

	^Text string: '' attributes: emphasisHere! !

!TextEditor methodsFor: 'attributes' stamp: 'jmv 7/27/2009 14:20'!
offerFontMenu
	"Present a menu of available fonts, and if one is chosen, apply it to the current selection.  
	Use only names of Fonts of this paragraph  "
	
	^self changeTextFont! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:44'!
offerFontMenu: aKeyboardEvent 
	"The user typed the command key that requests a font change; Offer the font menu.  5/27/96 sw"

	self closeTypeIn.
	self offerFontMenu.
	^ true! !

!TextEditor methodsFor: 'menu commands' stamp: 'jmv 9/2/2009 12:44'!
offerMenuFromEsc: aKeyboardEvent
	"The escape key was hit while the receiver has the keyboard focus; take action"

	^ ActiveEvent shiftPressed
		ifFalse: [
			self raiseContextMenu: aKeyboardEvent]! !

!TextEditor methodsFor: 'typing support' stamp: 'jmv 7/30/2009 00:46'!
openTypeIn
	"Set up UndoSelection to null text (to be added to by readKeyboard and backTo:),
	 beginTypeInBlock to keep track of the leftmost backspace, and UndoParameter to tally
	 how many deleted characters were backspaced over rather than 'cut'.
	 You can't undo typing until after closeTypeIn."

	beginTypeInBlock ifNil: [
		UndoSelection := self nullText.
		self undoer: #noUndoer with: 0.
		beginTypeInBlock := self startIndex]! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:45'!
outdent: aKeyboardEvent
	"Remove a tab from the front of every line occupied by the selection.
	Invoked from keyboard via cmd-shift-L.  2/29/96 sw"

	^ self inOutdent: aKeyboardEvent delta: -1! !

!TextEditor methodsFor: 'private' stamp: 'jmv 11/4/2008 11:51'!
pageHeight
	| howManyLines visibleHeight totalHeight ratio |
	howManyLines := paragraph numberOfLines.
	visibleHeight := self visibleHeight.
	totalHeight := self totalTextHeight.
	ratio := visibleHeight / totalHeight.
	^(ratio * howManyLines) rounded - 2! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
pasteRecent
	"Paste an item chose from RecentClippings."

	| clipping |
	(clipping := Clipboard chooseRecentClipping) ifNil: [^ self].
	Clipboard clipboardText: clipping.
	^ self paste! !

!TextEditor methodsFor: 'pluggable menus' stamp: 'jmv 9/2/2009 13:43'!
pluggableYellowButtonActivity: shiftKeyState
	"Invoke the model's popup menu."

	| menu |
	(menu := self getPluggableYellowButtonMenu: shiftKeyState)
		"ifNil: [
			sensor waitNoButton]"
		ifNotNil: [
			menu invokeOn: model orSendTo: self]! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:50'!
pointBlock
	^ pointBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 13:50'!
pointBlock: aCharacterBlock
	pointBlock := aCharacterBlock.
! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
pointIndex
	^ self pointBlock stringIndex! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 14:00'!
pointIndex: anIndex
	self pointBlock: (paragraph characterBlockForIndex: anIndex)
! !

!TextEditor methodsFor: 'typing support' stamp: 'jmv 9/2/2009 12:59'!
processKeyStroke: aKeyboardEvent
	"Key struck on the keyboard. Find out which one and, if special, carry 
	out the associated special action. Otherwise, add the character to the 
	stream of characters.  Undoer & Redoer: see closeTypeIn."
	
	
		self deselect.
		
		
			(self dispatchOn: aKeyboardEvent) ifTrue: [
				self doneTyping.
				self storeSelectionInParagraph.
				^self].
			self openTypeIn.
			
			
		self hasSelection ifTrue: [ "save highlighted characters"
			UndoSelection := self selection].
		self unselect.
		
		
	self storeSelectionInParagraph! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 12:44'!
raiseContextMenu: aKeyboardEvent

	(morph respondsTo: #editView)
		ifTrue: [morph editView yellowButtonActivity: ActiveEvent shiftPressed].
	^ true! !

!TextEditor methodsFor: 'current selection' stamp: 'jmv 11/4/2008 14:28'!
recomputeInterval
	"The same characters are selected but their coordinates may have changed."

	self markIndex: self markIndex; pointIndex: self pointIndex! !

!TextEditor methodsFor: 'current selection' stamp: 'jmv 4/6/2009 18:05'!
recomputeSelection
	"Redetermine the selection according to the start and stop block indices; 
	do not highlight."

	self deselect; recomputeInterval! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 11:51'!
replace: xoldInterval with: newText and: selectingBlock 
	"Replace the text in oldInterval with newText and execute selectingBlock to establish the new selection.  Create an undoAndReselect:redoAndReselect: undoer to allow perfect undoing."

	| undoInterval |
	undoInterval := self selectionInterval.
	undoInterval = xoldInterval ifFalse: [self selectInterval: xoldInterval].
	UndoSelection := self selection.
	self zapSelectionWith: newText.
	selectingBlock value.
	otherInterval := self selectionInterval.
	self undoer: #undoAndReselect:redoAndReselect: with: undoInterval with: otherInterval! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 11:51'!
replaceSelectionWith: aText
	"Remember the selection text in UndoSelection.
	 Deselect, and replace the selection text by aText.
	 Remember the resulting selectionInterval in UndoInterval and PriorInterval.
	 Set up undo to use UndoReplace."

	beginTypeInBlock ~~ nil ifTrue: [^self zapSelectionWith: aText]. "called from old code"
	UndoSelection := self selection.
	self zapSelectionWith: aText.
	self undoer: #undoReplace! !

!TextEditor methodsFor: 'initialize-release' stamp: 'jmv 9/1/2009 10:23'!
resetState 
	"Establish the initial conditions for editing the paragraph: place caret 
	before first character, set the emphasis to that of the first character,
	and save the paragraph for purposes of canceling."

	markBlock := paragraph defaultCharacterBlock.
	self pointBlock: markBlock copy.
	beginTypeInBlock := nil.
	UndoInterval := otherInterval := 1 to: 0.
	self setEmphasisHereFromText.
	selectionShowing := false! !

!TextEditor methodsFor: 'current selection' stamp: 'jmv 11/4/2008 11:51'!
reverseSelection
	"Reverse the valence of the current selection highlighting."
	selectionShowing := selectionShowing not.
	paragraph reverseFrom: self pointBlock to: self markBlock! !

!TextEditor methodsFor: 'private' stamp: 'AlainPlantec 4/15/2010 11:31'!
sameColumn: start newLine: lineBlock forward: isForward
	"Private - Compute the index in my text
	with the line number derived from lineBlock,"
	" a one argument block accepting the old line number.
	The position inside the line will be preserved as good as possible"
	"The boolean isForward is used in the border case to determine if
	we should move to the beginning or the end of the line."
	| column currentLine offsetAtTargetLine targetEOL lines numberOfLines currentLineNumber targetLineNumber |
	self walkAlongDisplayedLine
		ifTrue: [
			lines := paragraph lines.
			numberOfLines := paragraph numberOfLines.
			currentLineNumber  := paragraph lineIndexOfCharacterIndex: start.
			currentLine := lines at: currentLineNumber]
		ifFalse: [
			lines := self lines.
			numberOfLines := lines size.
			currentLine := lines
				detect:[:lineInterval | lineInterval last >= start]
				ifNone:[lines last].
			currentLineNumber := currentLine second].
	column := start - currentLine first.
	targetLineNumber := ((lineBlock value: currentLineNumber) max: 1) min: numberOfLines.
	offsetAtTargetLine := (lines at: targetLineNumber) first.
	targetEOL := (lines at: targetLineNumber) last + (targetLineNumber = numberOfLines ifTrue:[1]ifFalse:[0]).
	targetLineNumber = currentLineNumber
	"No movement or movement failed. Move to beginning or end of line."
		ifTrue:[^isForward
			ifTrue:[targetEOL]
			ifFalse:[offsetAtTargetLine]].
	^offsetAtTargetLine + column min: targetEOL.! !

!TextEditor methodsFor: 'menu messages' stamp: 'md 5/13/2010 13:02'!
saveContentsInFile
	"Save the receiver's contents string to a file, prompting the user for a file-name.  Suggest a reasonable file-name."

	| fileName stringToSave parentWindow labelToUse suggestedName lastIndex |
	stringToSave := paragraph text string. 
	stringToSave size = 0 ifTrue: [^self inform: 'nothing to save.'].
	parentWindow := model dependents 
				detect: [:dep | dep isKindOf: SystemWindow]
				ifNone: [nil].
	labelToUse := parentWindow ifNil: ['Untitled']
				ifNotNil: [parentWindow label].
	suggestedName := nil.
	#(#('Decompressed contents of: ' '.gz')) do: 
			[:leaderTrailer | 
			"can add more here..."

			(labelToUse beginsWith: leaderTrailer first) 
				ifTrue: 
					[suggestedName := labelToUse copyFrom: leaderTrailer first size + 1
								to: labelToUse size.
					(labelToUse endsWith: leaderTrailer last) 
						ifTrue: 
							[suggestedName := suggestedName copyFrom: 1
										to: suggestedName size - leaderTrailer last size]
						ifFalse: 
							[lastIndex := suggestedName lastIndexOf: $. ifAbsent: [0].
							(lastIndex = 0 or: [lastIndex = 1]) 
								ifFalse: [suggestedName := suggestedName copyFrom: 1 to: lastIndex - 1]]]].
	suggestedName ifNil: [suggestedName := labelToUse , '.text'].
	fileName := UIManager default request: 'File name?'
				initialAnswer: suggestedName.
	fileName isEmptyOrNil 
		ifFalse: 
			[(FileStream newFileNamed: fileName)
				nextPutAll: stringToSave;
				close]! !

!TextEditor methodsFor: 'scrolling' stamp: 'jmv 11/4/2008 11:51'!
scrollBy: ignore 
	"Ignore scroll requests."! !

!TextEditor methodsFor: 'typing/selecting keys' stamp: 'jmv 9/2/2009 13:27'!
search: aKeyboardEvent
	"Invoked by Ctrl-S.  Same as 'again', but always uses the existing FindText
	 and ChangeText regardless of the last edit."

	self closeTypeIn.
	self
		againOrSame: true "true means use same keys"
		many: aKeyboardEvent shiftPressed.
	^true! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 12:45'!
selectCurrentTypeIn: aKeyboardEvent 
	"Select what would be replaced by an undo (e.g., the last typeIn)."

	| prior |

	self closeTypeIn.
	prior := otherInterval.
	self closeTypeIn.
	self selectInterval: UndoInterval.
	otherInterval := prior.
	^ true! !

!TextEditor methodsFor: 'new selection' stamp: 'jmv 9/1/2009 10:23'!
selectFrom: start to: stop
	"Select the specified characters inclusive."
	self selectInvisiblyFrom: start to: stop.
	self closeTypeIn.
	self storeSelectionInParagraph.
	"Preserve current emphasis if selection is empty"
	stop > start ifTrue: [
		self setEmphasisHereFromText ]! !

!TextEditor methodsFor: 'new selection' stamp: 'FernandoOlivero 6/16/2010 12:53'!
selectLine
	"Make the receiver's selection, if it currently consists of an insertion point only, encompass the current line."
	self hasSelection ifTrue:[^self].
	self selectInterval: (self encompassParagraph: self selectionInterval)! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
selection
	"Answer the text in the paragraph that is currently selected."

	^paragraph text copyFrom: self startIndex to: self stopIndex - 1 ! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
selectionAsStream
	"Answer a ReadStream on the text in the paragraph that is currently 
	selected."

	^ReadStream
		on: (paragraph string copyFrom: self startIndex to: self stopIndex - 1)! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'StephaneDucasse 7/11/2010 15:51'!
selectionForDoitAsStream
	"Answer a ReadStream on the text in the paragraph that is currently 
	selected. Remove comment so that we can evaluate inside without removing them"

	^ReadStream
		on: (paragraph string withBlanksTrimmed withoutWrappingDoubleQuotes
				copyFrom: self startIndex 
				to: self stopIndex - 1)! !

!TextEditor methodsFor: 'typing support' stamp: 'jmv 9/1/2009 18:45'!
setEmphasisHereFromText

	self setEmphasisHereFromTextForward: true! !

!TextEditor methodsFor: 'typing support' stamp: 'FernandoOlivero 3/13/2010 15:40'!
setEmphasisHereFromTextForward: f

	| i t forward delta prevIsSeparator nextIsSeparator |
	i := self pointIndex.
	t := paragraph text.
	"Try to set emphasisHere correctly after whitespace.
	Most important after a cr, i.e. at the start of a new line"
	prevIsSeparator :=  i > 1 and: [ (t at: i-1) isSeparator ].
	nextIsSeparator := i <= t size and: [ (t at: i) isSeparator ].
	forward := prevIsSeparator = nextIsSeparator
		ifTrue: [ f ]
		ifFalse: [ nextIsSeparator ].
	delta := forward ifTrue: [ 1 ] ifFalse: [ 0 ].
	emphasisHere := (t attributesAt: (i - delta max: 1))
					select: [:att | att mayBeExtended].
	! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 11:51'!
setSearch: aString
	"Set the FindText and ChangeText to seek aString; except if already seeking aString, leave ChangeText alone so again will repeat last replacement."

	FindText string = aString
		ifFalse: [FindText := ChangeText := aString asText]! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 11/4/2008 11:51'!
setSearchString
	"Make the current selection, if any, be the current search string."
	self hasCaret ifTrue: [morph flash. ^ self].
	self setSearch:  self selection string! !

!TextEditor methodsFor: 'nonediting/nontyping keys' stamp: 'jmv 9/2/2009 12:45'!
setSearchString: aKeyboardEvent
	"Establish the current selection as the current search string."

	| aString |
	self closeTypeIn.
	self lineSelectAndEmptyCheck: [^ true].
	aString :=  self selection string.
	aString size = 0
		ifTrue:
			[self flash]
		ifFalse:
			[self setSearch: aString].
	^ true! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:45'!
shiftEnclose: aKeyboardEvent
	"Insert or remove bracket characters around the current selection."

	| char left right startIndex stopIndex oldSelection which text |
	char := aKeyboardEvent keyCharacter.
	char = $9 ifTrue: [ char := $( ].
	char = $, ifTrue: [ char := $< ].
	char = $[ ifTrue: [ char := ${ ].
	char = $' ifTrue: [ char := $" ].
	char asciiValue = 27 ifTrue: [ char := ${ ].	"ctrl-["

	self closeTypeIn.
	startIndex := self startIndex.
	stopIndex := self stopIndex.
	oldSelection := self selection.
	which := '([<{"''' indexOf: char ifAbsent: [1].
	left := '([<{"''' at: which.
	right := ')]>}"''' at: which.
	text := paragraph text.
	((startIndex > 1 and: [stopIndex <= text size])
			and: [ (text at: startIndex-1) = left and: [(text at: stopIndex) = right]])
		ifTrue: [
			"already enclosed; strip off brackets"
			self selectFrom: startIndex-1 to: stopIndex.
			self replaceSelectionWith: oldSelection]
		ifFalse: [
			"not enclosed; enclose by matching brackets"
			self replaceSelectionWith:
				(Text string: (String with: left), oldSelection string, (String with: right) attributes: emphasisHere).
			self selectFrom: startIndex+1 to: stopIndex].
	^true! !

!TextEditor methodsFor: 'pluggable menus' stamp: 'jmv 11/4/2008 11:51'!
shiftedTextPaneMenuRequest
	"The user chose the more... branch from the text-pane menu."

	^ self pluggableYellowButtonActivity: true! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
startBlock
	^ self pointBlock min: self markBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
startIndex
	^ self startBlock stringIndex! !

!TextEditor methodsFor: 'typing support' stamp: 'jmv 11/4/2008 11:51'!
startOfTyping
	"Compatibility during change from characterBlock to integer"
	beginTypeInBlock == nil ifTrue: [^ nil].
	beginTypeInBlock isNumber ifTrue: [^ beginTypeInBlock].
	"Last line for compatibility during change from CharacterBlock to Integer."
	^ beginTypeInBlock stringIndex! !

!TextEditor methodsFor: 'initialize-release' stamp: 'jmv 9/17/2009 16:06'!
stateArray
	^ {
		self selectionInterval.
		self startOfTyping.
		emphasisHere}! !

!TextEditor methodsFor: 'initialize-release' stamp: 'jmv 9/18/2009 15:48'!
stateArrayPut: stateArray
	| sel |
	sel := stateArray at: 1.
	self selectFrom: sel first to: sel last.
	beginTypeInBlock := stateArray at: 2.
	emphasisHere := stateArray at: 3.
	morph possiblyChanged! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
stopBlock
	^ self pointBlock max: self markBlock! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
stopIndex
	^ self stopBlock stringIndex! !

!TextEditor methodsFor: 'mvc compatibility' stamp: 'jmv 11/4/2008 11:51'!
storeSelectionInParagraph
	paragraph selectionStart: self startBlock selectionStop: self stopBlock! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 23:01'!
string

	^self text string! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:45'!
swapChars: aKeyboardEvent
	"Triggered byCmd-Y;.  Swap two characters, either those straddling the insertion point, or the two that comprise the selection.  Suggested by Ted Kaehler.  "

	| currentSelection aString chars |
	(chars := self selection) size = 0
		ifTrue:
			[currentSelection := self pointIndex.
			self selectMark: currentSelection - 1 point: currentSelection]
		ifFalse:
			[chars size = 2
				ifFalse:
					[morph flash. ^ true]
				ifTrue:
					[currentSelection := self pointIndex - 1]].
	aString := self selection string.
	self replaceSelectionWith: (Text string: aString reversed attributes: emphasisHere).
	self selectAt: currentSelection + 1.
	^ true! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 11:51'!
text
	"Answer the text of the paragraph being edited."

	^paragraph text! !

!TextEditor methodsFor: 'as yet unclassified' stamp: 'jmv 11/4/2008 11:51'!
totalTextHeight

	^paragraph lines last bottom! !

!TextEditor methodsFor: 'accessing' stamp: 'jmv 11/4/2008 11:51'!
transformFrom: owner
	^morph transformFrom: owner! !

!TextEditor methodsFor: 'private' stamp: 'jmv 9/21/2009 11:26'!
unapplyAttribute: aTextAttribute
	"The user selected aTextAttribute to be removed.
	If there is a selection, unapply the attribute to the selection.
	In any case do not use the attribute for the user input (emphasisHere)
	It really looks like the 2 branches in this method should be closer!! (handling of Undo is different, and likely wrong)"

	| thisSel interval |

	emphasisHere := emphasisHere copyWithout: aTextAttribute.

	aTextAttribute isParagraphAttribute
		ifTrue: [
			interval := paragraph text encompassParagraph: self selectionInterval.
			paragraph 
				replaceFrom: interval first 
				to: interval last 
				with: ((paragraph text copyFrom: interval first to: interval last)
						removeAttribute: aTextAttribute).
			paragraph composeAll.
			self recomputeSelection ]
		
		ifFalse: [
			thisSel := self selection.
			thisSel isEmpty
				ifTrue: [ paragraph composeAll ]
				ifFalse: [
					self replaceSelectionWith: (thisSel removeAttribute: aTextAttribute).
					paragraph composeAll.
					self recomputeSelection] ].
	morph possiblyChanged ! !

!TextEditor methodsFor: 'private' stamp: 'jmv 9/18/2009 15:48'!
unapplyAttributesThat: removalBlock fromWholeParagraph: wholeParagraph
	"The user selected aTextAttribute to be removed.
	If there is a selection, unapply the attribute to the selection.
	In any case do not use the attribute for the user input (emphasisHere)
	It really looks like the 2 branches in this method should be closer!! (handling of Undo is different, and likely wrong)"

	| interval newText secondPass |

	emphasisHere := emphasisHere reject: removalBlock.
	wholeParagraph
		ifTrue: [
			interval := paragraph text encompassParagraph: self selectionInterval.
			newText := paragraph text copyFrom: interval first to: interval last.
			newText
				removeAttributesThat: removalBlock
				replaceAttributesThat: [ :attr | false ]
				by: [ :attr | false ]
				signalIfWholeParagraphAttribute: false.
			paragraph 
				replaceFrom: interval first 
				to: interval last 
				with: newText.
			paragraph composeAll.
			self recomputeSelection ]
		
		ifFalse: [
			newText := self selection.
			newText notEmpty
				ifTrue: [
					"Do a first pass that will remove only regular attributes (i.e. isParagraphAttribute not)"
					secondPass := false.
					[
						newText
							removeAttributesThat: removalBlock
							replaceAttributesThat: [ :attr | false ]
							by: [ :attr | false ]
							signalIfWholeParagraphAttribute: true
					] on: Abort do: [ :ex |
						secondPass := true ].
					self replaceSelectionWith: newText.
					paragraph composeAll.
					self recomputeSelection.
					"Do a second pass, calling this method again, that will remove paragraph attributes from whole paragraphs"
					secondPass ifTrue: [
						self unapplyAttributesThat: [ :attr | attr isParagraphAttribute and: [removalBlock value: attr ]] fromWholeParagraph: true ] ] ].
	morph possiblyChanged ! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 9/2/2009 13:31'!
undo
	"Reset the state of the paragraph prior to the previous edit.
	 If another ParagraphEditor instance did that edit, UndoInterval is invalid;
	 just recover the contents of the undo-buffer at the start of the paragraph."

	self closeTypeIn.

	UndoParagraph == paragraph ifFalse: "Can't undo another paragraph's edit"
		[UndoMessage := Message selector: #undoReplace.
		UndoInterval := 1 to: 0.
		Undone := true].
	UndoInterval ~= self selectionInterval ifTrue: "blink the actual target"
		[self selectInterval: UndoInterval; deselect].

	"Leave a signal of which phase is in progress"
	UndoParagraph := Undone ifTrue: [#redoing] ifFalse: [#undoing].
	UndoMessage sentTo: self.
	UndoParagraph := paragraph! !

!TextEditor methodsFor: 'editing keys' stamp: 'jmv 9/2/2009 12:45'!
undo: aKeyboardEvent 
	"Undo the last edit."

	self closeTypeIn.
	self undo.
	^true! !

!TextEditor methodsFor: 'undoers' stamp: 'jmv 11/4/2008 11:51'!
undoAgain: indices andReselect: home typedKey: wasTypedKey
	"The last command was again.  Undo it. Redoer: itself."

	| findSize substText index subject |
	(self isRedoing & wasTypedKey) ifTrue: "redelete search key"
		[self selectInterval: home.
		self zapSelectionWith: self nullText].

	findSize := (self isRedoing ifTrue: [FindText] ifFalse: [ChangeText]) size.
	substText := self isUndoing ifTrue: [FindText] ifFalse: [ChangeText].
	(self isUndoing ifTrue: [indices size to: 1 by: -1] ifFalse: [1 to: indices size]) do:
		[:i |
		index := indices at: i.
		(subject := index to: index + findSize - 1) = self selectionInterval ifFalse:
			[self selectInterval: subject].
		FindText == ChangeText ifFalse: [self zapSelectionWith: substText]].

	self isUndoing
		ifTrue:  "restore selection to where it was when 'again' was invoked"
			[wasTypedKey
				ifTrue: "search started by typing key at a caret; restore it"
					[self selectAt: home first.
					self zapSelectionWith: FindText.
					self selectAt: home last + 1]
				ifFalse: [self selectInterval: home]].

	self undoMessage: UndoMessage forRedo: self isUndoing! !

!TextEditor methodsFor: 'undoers' stamp: 'jmv 11/4/2008 11:51'!
undoAndReselect: undoHighlight redoAndReselect: redoHighlight
	"Undo typing, cancel, paste, and other operations that are like replaces
	 but the selection is not the whole restored text after undo, redo, or both.
	 undoHighlight is selected after this phase and redoHighlight after the next phase.
	Redoer: itself."

	self replace: self selectionInterval with: UndoSelection and:
		[self selectInterval: undoHighlight].
	self undoMessage: (UndoMessage argument: redoHighlight) forRedo: self isUndoing
! !

!TextEditor methodsFor: 'undoers' stamp: 'jmv 11/4/2008 11:51'!
undoCutCopy: oldPasteBuffer
	"Undo of a cut, copy, or any edit that changed CurrentSelection.  Be sure
	 undo-copy does not lock the model.  Redoer: itself, so never isRedoing."

	| recentCut |
	recentCut := self clipboardText.	
	UndoSelection size = UndoInterval size
		ifFalse: [self replaceSelectionWith: UndoSelection].
	self clipboardTextPut: oldPasteBuffer.
	self undoer: #undoCutCopy: with: recentCut! !

!TextEditor methodsFor: 'undo support' stamp: 'jmv 11/4/2008 11:51'!
undoMessage: aMessage forRedo: aBoolean
	"Call this from an undoer/redoer to set up UndoMessage as the
	 corresponding redoer/undoer.  Also set up UndoParagraph, as well
	 as the state variable Undone.  It is assumed that UndoInterval has been
	 established (generally by zapSelectionWith:) and that UndoSelection has been
	 saved (generally by replaceSelectionWith: or replace:With:and:)."

	self isDoing ifTrue: [UndoParagraph := paragraph].
	UndoMessage := aMessage.
	Undone := aBoolean! !

!TextEditor methodsFor: 'undoers' stamp: 'jmv 11/4/2008 11:51'!
undoReplace
	"Undo of any command that replaced a selection by other text that it left
	 highlighted, and that is undone and redone by simple reversal of the
	 operation.  This is the most common Undoer; call replaceSelectionWith:
	 to get this setup.  Redoer: itself, so never isRedoing."

	self replaceSelectionWith: UndoSelection! !

!TextEditor methodsFor: 'undo support' stamp: 'jmv 11/4/2008 11:51'!
undoer: aSelector
	"See comment in undoMessage:.  Use this version when aSelector has no arguments, and you are doing or redoing and want to prepare for undoing."

	self undoMessage: (Message selector: aSelector) forRedo: false! !

!TextEditor methodsFor: 'undo support' stamp: 'jmv 11/4/2008 11:51'!
undoer: aSelector with: arg1
	"See comment in undoMessage:.  Use this version when aSelector has one argument, and you are doing or redoing and want to prepare for undoing."

	self undoMessage: (Message selector: aSelector argument: arg1) forRedo: false! !

!TextEditor methodsFor: 'undo support' stamp: 'jmv 11/4/2008 11:51'!
undoer: aSelector with: arg1 with: arg2
	"See comment in undoMessage:.  Use this version when aSelector has two arguments, and you are doing or redoing and want to prepare for undoing."

	self undoMessage: (Message selector: aSelector arguments: (Array with: arg1 with: arg2)) forRedo: false! !

!TextEditor methodsFor: 'undo support' stamp: 'jmv 11/4/2008 11:51'!
undoer: aSelector with: arg1 with: arg2 with: arg3
	"See comment in undoMessage:.  Use this version when aSelector has three arguments, and you are doing or redoing and want to prepare for undoing."

	self undoMessage: (Message selector: aSelector arguments: (Array with: arg1 with: arg2 with: arg3)) forRedo: false! !

!TextEditor methodsFor: 'accessing-selection' stamp: 'jmv 11/4/2008 11:51'!
unselect
	self markBlock: self pointBlock copy! !

!TextEditor methodsFor: 'as yet unclassified' stamp: 'FernandoOlivero 3/13/2010 16:23'!
visibleHeight

	^morph owner bounds height
! !

!TextEditor methodsFor: 'settings' stamp: 'AlainPlantec 4/15/2010 11:28'!
walkAlongDisplayedLine
	^ self class walkAlongDisplayedLine! !

!TextEditor methodsFor: 'menu messages' stamp: 'jmv 9/2/2009 13:44'!
yellowButtonActivity
	"This normally opens a popup menu. Determine the selected
	item and, if one is selected, then send the corresponding message
	to either the model or the receiver."

	^ self pluggableYellowButtonActivity: false! !

!TextEditor methodsFor: 'mvc compatibility' stamp: 'FernandoOlivero 3/13/2010 16:02'!
zapSelectionWith: aText

	| start stop |
	self deselect.
	start := self startIndex.
	stop := self stopIndex.
	(aText isEmpty and: [stop > start]) ifTrue: [
		"If deleting, then set emphasisHere from 1st character of the deletion"
		emphasisHere := (paragraph text attributesAt: start) select: [:att | att mayBeExtended]].
	(start = stop and: [ aText size = 0 ]) ifFalse: [
		paragraph replaceFrom: start to: stop - 1 with: aText.
		self markIndex: start; pointIndex: start + aText size.
		UndoInterval := otherInterval := self selectionInterval.

		self userHasEdited  " -- note text now dirty" ].
! !
TextEditor initialize!
SimpleEditor initialize!
